from bloqade.submission.ir.task_specification import QuEraTaskSpecification
import bloqade.submission.quera_api_client.api
from bloqade.submission.base import get_capabilities
from unittest.mock import patch
from requests import Response
import simplejson as json
from typing import Dict

from bloqade import start
from bloqade.task.batch import RemoteBatch
import pytest
import os


API_HOSTNAME = "api.que-ee.com"
API_PROXY = "proxy-api.que-ee.com"
VIRTUAL_QUEUE = "virtual-queue"

API_CONFIG = dict(
    api_hostname=API_HOSTNAME,
    qpu_id="qpu-1",
    api_stage="v0",
    proxy=API_PROXY,
    virtual_queue=VIRTUAL_QUEUE,
)

HEADERS = {
    "Content-Type": "application/json",
    "Host": API_HOSTNAME,
    "virtual_queue": VIRTUAL_QUEUE,
}


# Integraiton tests
@pytest.mark.skip(reason="Depricating save_batch")
@pytest.mark.vcr
def test_quera_submit():
    os.environ["AWS_ACCESS_KEY"] = "XXXXXXXXXXXXXXXXXXXX"
    os.environ["AWS_SECRET_KEY"] = "XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX"
    os.environ["AWS_SESSION_TOKEN"] = 900 * "X"
    # credentials = {
    #     "access_key": "XXXXXXXXXXXXXXXXXXXX",
    #     "secret_key": "XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX",
    #     "session_token": 900 * "X",
    # }
    config_file = os.path.join("tests", "data", "config", "submit_quera_api.json")

    batch = (
        start.add_position((0, 0))
        .rydberg.rabi.amplitude.uniform.piecewise_linear(
            [0.1, "run_time", 0.1], [0, 15, 15, 0]
        )
        .assign(run_time=2.0)
        .parallelize(20)
        .quera.device(config_file=config_file)
        .run_async(shots=10)
    )

    bloqade.save_batch("quera_submit.json", batch)


@pytest.mark.skip(reason="Depricating load_batch")
@pytest.mark.vcr
def test_quera_retrieve():
    from bloqade.task.json import load_batch

    credentials = {
        "access_key": "XXXXXXXXXXXXXXXXXXXX",
        "secret_key": "XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX",
        "session_token": 900 * "X",
    }

    job_future = load_batch("quera_submit.json", **credentials)
    assert type(job_future) == RemoteBatch
    job_future.pull()


def create_response(
    status_code: int, content: Dict, content_type="application/json"
) -> Response:
    response = Response()
    response.status_code = status_code
    response.headers["Content-Type"] = content_type
    response._content = bytes(json.dumps(content), "utf-8")
    return response


def mock_task_ir():
    return {
        "nshots": 10,
        "lattice": {"sites": [(0, 0)], "filling": [1]},
        "effective_hamiltonian": {
            "rydberg": {
                "rabi_frequency_amplitude": {
                    "global": {
                        "times": [0, 1e-6, 2e-6, 3e-6, 4e-6],
                        "values": [0, 15e6, 15e6, 0],
                    },
                },
                "rabi_frequency_phase": {
                    "global": {"times": [0, 4e-6], "values": [0, 0]},
                },
                "detuning": {
                    "global": {
                        "times": [0, 1e-6, 2e-6, 3e-6, 4e-6],
                        "values": [0, 15e6, 15e6, 0],
                    },
                },
            }
        },
    }


def mock_results():
    return {
        "task_status": "Completed",
        "shot_outputs": [
            {"shot_status": "Completed", "pre_sequence": [1], "post_sequence": [1]},
            {"shot_status": "Completed", "pre_sequence": [1], "post_sequence": [1]},
            {"shot_status": "Completed", "pre_sequence": [1], "post_sequence": [1]},
            {"shot_status": "Completed", "pre_sequence": [1], "post_sequence": [1]},
            {"shot_status": "Completed", "pre_sequence": [1], "post_sequence": [1]},
            {"shot_status": "Completed", "pre_sequence": [1], "post_sequence": [1]},
            {"shot_status": "Completed", "pre_sequence": [1], "post_sequence": [1]},
            {"shot_status": "Completed", "pre_sequence": [1], "post_sequence": [1]},
            {"shot_status": "Completed", "pre_sequence": [1], "post_sequence": [1]},
            {"shot_status": "Completed", "pre_sequence": [1], "post_sequence": [1]},
        ],
    }


@patch("bloqade.submission.quera_api_client.api.Sigv4Request")
def test_happy_path_queue_api(*args):
    def mock_get(*args, **kwargs):
        print("get", args, kwargs)
        if (args, kwargs) == (
            ("https://" + API_PROXY + "/v0/qpu-1/capabilities",),
            {"headers": HEADERS},
        ):
            return create_response(200, get_capabilities().dict())
        elif (args, kwargs) == (
            ("https://" + API_PROXY + "/v0/qpu-1/queue/task/task_id",),
            {"headers": HEADERS},
        ):
            return create_response(200, {"status": "Completed"})
        elif (args, kwargs) == (
            ("https://" + API_PROXY + "/v0/qpu-1/task/task_id/results",),
            {"headers": HEADERS},
        ):
            return create_response(200, mock_results())
        else:
            raise NotImplementedError

    def mock_put(*args, **kwargs):
        print("put", args, kwargs)
        if (args, kwargs) == (
            ("https://" + API_PROXY + "/v0/qpu-1/queue/task/task_id/cancel",),
            {"headers": HEADERS},
        ):
            return create_response(200, {})
        else:
            raise NotImplementedError

    def mock_post(*args, **kwargs):
        print("post", args, kwargs)
        if (
            args == ("https://" + API_PROXY + "/v0/qpu-1/queue/task",)
            and kwargs["headers"]["Content-Type"] == "application/json"
        ):
            task_ir_string = kwargs["data"]

            if QuEraTaskSpecification(
                **json.loads(task_ir_string)
            ) != QuEraTaskSpecification(**mock_task_ir()):
                raise ValueError("Task IRs do not match")

            return create_response(201, {"task_id": "task_id"})
        else:
            raise NotImplementedError

    sig_v4_config = dict(
        region="us-east-1",
        access_key=None,
        secret_key=None,
        session_token=None,
        session_expires=3600,
        role_arn=None,
        role_session_name="awsrequest",
        profile=None,
    )

    request_api = bloqade.submission.quera_api_client.api.Sigv4Request(**sig_v4_config)
    request_api.get.side_effect = mock_get
    request_api.post.side_effect = mock_post
    request_api.put.side_effect = mock_put
    queue = bloqade.submission.quera_api_client.api.QueueApi(**API_CONFIG)

    assert queue.get_capabilities() == get_capabilities().dict()
    assert queue.post_task(mock_task_ir()) == "task_id"
    assert queue.get_task_status_in_queue("task_id") == "Completed"
    assert queue.get_task_results("task_id") == mock_results()
    assert queue.is_task_stopped("task_id")
    assert queue.cancel_task_in_queue("task_id") is None


@patch("bloqade.submission.quera_api_client.api.Sigv4Request")
def test_task_results_polling(*args):
    sig_v4_config = dict(
        region="us-east-1",
        access_key=None,
        secret_key=None,
        session_token=None,
        session_expires=3600,
        role_arn=None,
        role_session_name="awsrequest",
        profile=None,
    )

    request_api = bloqade.submission.quera_api_client.api.Sigv4Request(**sig_v4_config)
    queue = bloqade.submission.quera_api_client.api.QueueApi(**API_CONFIG)

    request_api.get.side_effect = [
        create_response(200, {"status": "Enqueued"}),
        create_response(200, {"status": "Enqueued"}),
        create_response(200, {"status": "Enqueued"}),
        create_response(200, {"status": "Accepted"}),
        create_response(200, {"status": "Created"}),
        create_response(200, {"status": "Running"}),
        create_response(200, {"status": "Completed"}),
        # calls get_task_status_in_queue before fetching results
        create_response(200, {"status": "Completed"}),
        create_response(200, mock_results()),
    ]

    assert queue.poll_task_results("task_id", polling_interval=0.5) == mock_results()


@patch("bloqade.submission.quera_api_client.api.Sigv4Request")
def test_post_task_error_paths(*args):
    sig_v4_config = dict(
        region="us-east-1",
        access_key=None,
        secret_key=None,
        session_token=None,
        session_expires=3600,
        role_arn=None,
        role_session_name="awsrequest",
        profile=None,
    )

    QueueApi = bloqade.submission.quera_api_client.api.QueueApi
    request_api = bloqade.submission.quera_api_client.api.Sigv4Request(**sig_v4_config)
    queue = bloqade.submission.quera_api_client.api.QueueApi(**API_CONFIG)

    request_api.post.side_effect = [
        create_response(404, {}),
        create_response(400, {}),
        create_response(403, {}),
        create_response(500, {}),
        create_response(201, {"task_id_other": "task_id"}),
    ]

    with pytest.raises(QueueApi.NotFound):
        queue.post_task(mock_task_ir())

    with pytest.raises(QueueApi.InvalidRequestError):
        queue.post_task(mock_task_ir())

    with pytest.raises(QueueApi.AuthenticationError):
        queue.post_task(mock_task_ir())

    with pytest.raises(QueueApi.QueueApiError):
        queue.post_task(mock_task_ir())

    with pytest.raises(QueueApi.InvalidResponseError):
        queue.post_task(mock_task_ir())


@patch("bloqade.submission.quera_api_client.api.Sigv4Request")
def test_task_validation_error_paths(*args):
    sig_v4_config = dict(
        region="us-east-1",
        access_key=None,
        secret_key=None,
        session_token=None,
        session_expires=3600,
        role_arn=None,
        role_session_name="awsrequest",
        profile=None,
    )

    QueueApi = bloqade.submission.quera_api_client.api.QueueApi
    request_api = bloqade.submission.quera_api_client.api.Sigv4Request(**sig_v4_config)
    queue = bloqade.submission.quera_api_client.api.QueueApi(**API_CONFIG)

    request_api.post.side_effect = [
        create_response(404, {}),
        create_response(400, {}),
        create_response(403, {}),
        create_response(500, {}),
    ]

    with pytest.raises(QueueApi.NotFound):
        queue.validate_task(mock_task_ir())

    with pytest.raises(QueueApi.ValidationError):
        queue.validate_task(mock_task_ir())

    with pytest.raises(QueueApi.AuthenticationError):
        queue.validate_task(mock_task_ir())

    with pytest.raises(QueueApi.QueueApiError):
        queue.validate_task(mock_task_ir())


@patch("bloqade.submission.quera_api_client.api.Sigv4Request")
def test_get_capabilities_error_paths(*args):
    sig_v4_config = dict(
        region="us-east-1",
        access_key=None,
        secret_key=None,
        session_token=None,
        session_expires=3600,
        role_arn=None,
        role_session_name="awsrequest",
        profile=None,
    )

    QueueApi = bloqade.submission.quera_api_client.api.QueueApi
    request_api = bloqade.submission.quera_api_client.api.Sigv4Request(**sig_v4_config)
    queue = bloqade.submission.quera_api_client.api.QueueApi(**API_CONFIG)

    request_api.get.side_effect = [
        create_response(404, {}),
        create_response(403, {}),
        create_response(500, {}),
    ]

    with pytest.raises(QueueApi.NotFound):
        queue.get_capabilities()

    with pytest.raises(QueueApi.AuthenticationError):
        queue.get_capabilities()

    with pytest.raises(QueueApi.QueueApiError):
        queue.get_capabilities()


def test_api_requests_errors():
    api_requests = bloqade.submission.quera_api_client.api.ApiRequest(**API_CONFIG)

    with pytest.raises(NotImplementedError):
        api_requests._post("https://api.que-ee.com/v0/qpu-1", {}, {})

    with pytest.raises(NotImplementedError):
        api_requests._put("https://api.que-ee.com/v0/qpu-1", {})

    with pytest.raises(NotImplementedError):
        api_requests._get("https://api.que-ee.com/v0/qpu-1", {})


@patch("bloqade.submission.quera_api_client.api.Sigv4Request")
def test_get_task_status_in_queue_error_paths(*args):
    sig_v4_config = dict(
        region="us-east-1",
        access_key=None,
        secret_key=None,
        session_token=None,
        session_expires=3600,
        role_arn=None,
        role_session_name="awsrequest",
        profile=None,
    )

    QueueApi = bloqade.submission.quera_api_client.api.QueueApi
    request_api = bloqade.submission.quera_api_client.api.Sigv4Request(**sig_v4_config)
    queue = bloqade.submission.quera_api_client.api.QueueApi(**API_CONFIG)

    request_api.get.side_effect = [
        create_response(400, {}),
        create_response(404, {}),
        create_response(500, {}),
    ]

    with pytest.raises(QueueApi.InvalidRequestError):
        queue.get_task_status_in_queue("task_id")

    with pytest.raises(QueueApi.NotFound):
        queue.get_task_status_in_queue("task_id")

    with pytest.raises(QueueApi.QueueApiError):
        queue.get_task_status_in_queue("task_id")


@patch("bloqade.submission.quera_api_client.api.Sigv4Request")
def test_cancel_task_in_queue_error_paths(*args):
    sig_v4_config = dict(
        region="us-east-1",
        access_key=None,
        secret_key=None,
        session_token=None,
        session_expires=3600,
        role_arn=None,
        role_session_name="awsrequest",
        profile=None,
    )

    QueueApi = bloqade.submission.quera_api_client.api.QueueApi
    request_api = bloqade.submission.quera_api_client.api.Sigv4Request(**sig_v4_config)
    queue = bloqade.submission.quera_api_client.api.QueueApi(**API_CONFIG)

    request_api.put.side_effect = [
        create_response(403, {}),
        create_response(404, {}),
        create_response(500, {}),
    ]

    with pytest.raises(QueueApi.AuthenticationError):
        queue.cancel_task_in_queue("task_id")

    with pytest.raises(QueueApi.NotFound):
        queue.cancel_task_in_queue("task_id")

    with pytest.raises(QueueApi.QueueApiError):
        queue.cancel_task_in_queue("task_id")


@patch("bloqade.submission.quera_api_client.api.Sigv4Request")
def test_get_task_results_paths(*args):
    from bloqade.submission.quera_api_client.api import ApiRequest

    sig_v4_config = dict(
        region="us-east-1",
        access_key=None,
        secret_key=None,
        session_token=None,
        session_expires=3600,
        role_arn=None,
        role_session_name="awsrequest",
        profile=None,
    )

    QueueApi = bloqade.submission.quera_api_client.api.QueueApi
    request_api = bloqade.submission.quera_api_client.api.Sigv4Request(**sig_v4_config)
    queue = bloqade.submission.quera_api_client.api.QueueApi(**API_CONFIG)

    request_api.get.side_effect = [
        create_response(200, {"status": "Created"}),
        create_response(200, {"status": "Enqueued"}),
        create_response(200, {"status": "Accepted"}),
        # ----------------------------------------- #
        create_response(200, {"status": "Executing"}),
        # ----------------------------------------- #
        create_response(200, {"status": "Failed"}),
        create_response(200, {"status": "Cancelled"}),
        # ----------------------------------------- #
        create_response(200, {"status": "Unaccepted"}),
        # ----------------------------------------- #
        create_response(200, {"status": "OtherRandomMessage"}),
        # ----------------------------------------- #
        create_response(200, {"status": "Completed"}),
        create_response(400, {}),
        # ----------------------------------------- #
        create_response(200, {"status": "Completed"}),
        create_response(404, {}),
        # ----------------------------------------- #
        create_response(200, {"status": "Completed"}),
        create_response(403, {}),
        # ----------------------------------------- #
        create_response(200, {"status": "Completed"}),
        create_response(500, {}),
        # ----------------------------------------- #
        create_response(200, {"status": "Completed"}),
        create_response(200, {"status": "Completed"}, content_type="text/html"),
    ]

    assert queue.get_task_results("task_id") == {
        "task_status": "Created",
        "shot_outputs": [],
    }
    assert queue.get_task_results("task_id") == {
        "task_status": "Created",
        "shot_outputs": [],
    }
    assert queue.get_task_results("task_id") == {
        "task_status": "Created",
        "shot_outputs": [],
    }
    assert queue.get_task_results("task_id") == {
        "task_status": "Running",
        "shot_outputs": [],
    }
    assert queue.get_task_results("task_id") == {
        "task_status": "Failed",
        "shot_outputs": [],
    }
    assert queue.get_task_results("task_id") == {
        "task_status": "Cancelled",
        "shot_outputs": [],
    }

    with pytest.raises(QueueApi.ValidationError):
        queue.get_task_results("task_id")

    with pytest.raises(QueueApi.QueueApiError):
        queue.get_task_results("task_id")

    with pytest.raises(QueueApi.InvalidRequestError):
        queue.get_task_results("task_id")

    with pytest.raises(QueueApi.NotFound):
        queue.get_task_results("task_id")

    with pytest.raises(QueueApi.AuthenticationError):
        queue.get_task_results("task_id")

    with pytest.raises(QueueApi.QueueApiError):
        queue.get_task_results("task_id")

    with pytest.raises(ApiRequest.InvalidResponseError):
        queue.get_task_results("task_id")


@patch("bloqade.submission.quera_api_client.api.Sigv4Request")
def test_get_task_summary_paths(*args):
    sig_v4_config = dict(
        region="us-east-1",
        access_key=None,
        secret_key=None,
        session_token=None,
        session_expires=3600,
        role_arn=None,
        role_session_name="awsrequest",
        profile=None,
    )

    QueueApi = bloqade.submission.quera_api_client.api.QueueApi
    request_api = bloqade.submission.quera_api_client.api.Sigv4Request(**sig_v4_config)
    queue = bloqade.submission.quera_api_client.api.QueueApi(**API_CONFIG)

    request_api.get.side_effect = [
        create_response(200, {"status": "Running"}),
        # ----------------------------------------- #
        create_response(200, {"status": "Completed"}),
        create_response(200, {"status": "Completed"}),
        # ----------------------------------------- #
        create_response(200, {"status": "Completed"}),
        create_response(404, {}),
        # ----------------------------------------- #
        create_response(200, {"status": "Completed"}),
        create_response(403, {}),
        # ----------------------------------------- #
        create_response(200, {"status": "Completed"}),
        create_response(500, {}),
        # ----------------------------------------- #
    ]

    with pytest.raises(QueueApi.QueueApiError):
        queue.get_task_summary("task_id")

    assert queue.get_task_summary("task_id") == {"status": "Completed"}

    with pytest.raises(QueueApi.NotFound):
        queue.get_task_summary("task_id")

    with pytest.raises(QueueApi.AuthenticationError):
        queue.get_task_summary("task_id")

    with pytest.raises(QueueApi.QueueApiError):
        queue.get_task_summary("task_id")



import bloqade.ir.control.waveform as waveform
from bloqade.ir.visitor.waveform import WaveformVisitor

from typing import Dict, Tuple, List, Union
from pydantic.dataclasses import dataclass
from bisect import bisect_left
import numbers
from decimal import Decimal


@dataclass(frozen=True)
class PiecewiseConstant:
    times: List[Decimal]
    values: List[Decimal]

    def eval(self, time):
        if time >= self.times[-1]:
            return self.values[-1]
        elif time <= self.times[0]:
            return self.values[0]
        else:
            index = bisect_left(self.times, time)

            if self.times[index] == time:
                index += 1

            return self.values[index]

    def slice(self, start_time, stop_time) -> "PiecewiseConstant":
        start_time = Decimal(str(start_time))
        stop_time = Decimal(str(stop_time))

        if start_time == stop_time:
            return PiecewiseConstant(
                [Decimal(0.0), Decimal(0.0)], [Decimal(0.0), Decimal(0.0)]
            )

        start_index = bisect_left(self.times, start_time)
        stop_index = bisect_left(self.times, stop_time)

        if start_time == self.times[start_index]:
            if stop_time == self.times[stop_index]:
                absolute_times = list(self.times[start_index : stop_index + 1])
                values = list(self.values[start_index : stop_index + 1])
            else:
                absolute_times = self.times[start_index:stop_index] + [stop_time]
                values = self.values[start_index:stop_index] + [self.values[stop_index]]
        else:
            if stop_time == self.times[stop_index]:
                absolute_times = [start_time] + self.times[start_index : stop_index + 1]
                values = [self.values[start_index - 1]] + self.values[
                    start_index : stop_index + 1
                ]
            else:
                absolute_times = (
                    [start_time] + self.times[start_index:stop_index] + [stop_time]
                )
                values = (
                    [self.values[start_index - 1]]
                    + self.values[start_index:stop_index]
                    + [self.values[stop_index]]
                )

        values[-1] = values[-2]

        return PiecewiseConstant([time - start_time for time in absolute_times], values)

    def append(self, other: "PiecewiseConstant"):
        return PiecewiseConstant(
            times=self.times + [time + self.times[-1] for time in other.times[1:]],
            values=self.values[:-1] + other.values,
        )


class PiecewiseConstantCodeGen(WaveformVisitor):
    def __init__(self, assignments: Dict[str, Union[numbers.Real, List[numbers.Real]]]):
        self.assignments = assignments
        self.times = []
        self.values = []

    def append_timeseries(self, value, duration):
        if len(self.times) > 0:
            self.times.append(duration + self.times[-1])
            self.values[-1] = value
            self.values.append(value)
        else:
            self.times = [Decimal(0), duration]
            self.values = [value, value]

    def visit_linear(self, ast: waveform.Linear) -> Tuple[List[Decimal], List[Decimal]]:
        duration = ast.duration(**self.assignments)
        start = ast.start(**self.assignments)
        stop = ast.stop(**self.assignments)

        if start != stop:
            raise ValueError(
                "Failed to compile Waveform to piecewise constant, "
                "found non-constant Linear piece."
            )

        self.append_timeseries(start, duration)

    def visit_constant(
        self, ast: waveform.Constant
    ) -> Tuple[List[Decimal], List[Decimal]]:
        duration = ast.duration(**self.assignments)
        value = ast.value(**self.assignments)

        self.append_timeseries(value, duration)

    def visit_poly(self, ast: waveform.Poly) -> Tuple[List[Decimal], List[Decimal]]:
        order = len(ast.coeffs) - 1
        duration = ast.duration(**self.assignments)

        if len(ast.coeffs) == 0:
            value = Decimal(0)

        elif len(ast.coeffs) == 1:
            value = ast.coeffs[0](**self.assignments)

        elif len(ast.coeffs) == 2:
            start = ast.coeffs[0](**self.assignments)
            stop = start + ast.coeffs[1](**self.assignments) * duration

            if start != stop:
                raise ValueError(
                    "Failed to compile Waveform to piecewise constant, "
                    "found non-constant Polynomial piece."
                )

            value = start

        else:
            raise ValueError(
                "Failed to compile Waveform to piecewise constant,"
                f"found Polynomial of order {order}."
            )

        self.append_timeseries(value, duration)

    def visit_negative(
        self, ast: waveform.Negative
    ) -> Tuple[List[Decimal], List[Decimal]]:
        self.visit(ast.waveform)

        self.values = [-value for value in self.values]

    def visit_scale(self, ast: waveform.Scale) -> Tuple[List[Decimal], List[Decimal]]:
        self.visit(ast.waveform)
        scale = ast.scalar(**self.assignments)
        self.values = [scale * value for value in self.values]

    def visit_slice(self, ast: waveform.Slice) -> Tuple[List[Decimal], List[Decimal]]:
        duration = ast.waveform.duration(**self.assignments)

        if ast.interval.start is None:
            start_time = Decimal(0)
        else:
            start_time = ast.interval.start(**self.assignments)

        if ast.interval.stop is None:
            stop_time = duration
        else:
            stop_time = ast.interval.stop(**self.assignments)

        if start_time < 0:
            raise ValueError((f"start time for slice {start_time} is smaller than 0."))

        if stop_time > duration:
            raise ValueError(
                (f"end time for slice {stop_time} is larger than duration {duration}.")
            )

        if stop_time < start_time:
            raise ValueError(
                (
                    f"end time for slice {stop_time} is smaller than "
                    f"starting value for slice {start_time}."
                )
            )

        new_pwc = (
            PiecewiseConstantCodeGen(self.assignments)
            .emit(ast.waveform)
            .slice(start_time, stop_time)
        )

        self.times = new_pwc.times
        self.values = new_pwc.values

    def visit_append(self, ast: waveform.Append) -> Tuple[List[Decimal], List[Decimal]]:
        pwc = PiecewiseConstantCodeGen(self.assignments).emit(ast.waveforms[0])

        for sub_expr in ast.waveforms[1:]:
            new_pwc = PiecewiseConstantCodeGen(self.assignments).emit(sub_expr)

            # skip instructions with duration=0
            if new_pwc.times[-1] == Decimal(0):
                continue

            pwc = pwc.append(new_pwc)

        self.times = pwc.times
        self.values = pwc.values

    def visit_sample(self, ast: waveform.Sample) -> Tuple[List[Decimal], List[Decimal]]:
        if ast.interpolation is not waveform.Interpolation.Constant:
            raise ValueError(
                "Failed to compile waveform to piecewise constant, "
                f"found piecewise {ast.interpolation.value} interpolation."
            )
        self.times, values = ast.samples(**self.assignments)
        values[-1] = values[-2]
        self.values = values

    def visit_record(self, ast: waveform.Record) -> Tuple[List[Decimal], List[Decimal]]:
        self.visit(ast.waveform)

    def emit(self, ast: waveform.Waveform) -> PiecewiseConstant:
        self.visit(ast)

        return PiecewiseConstant(times=self.times, values=self.values)



from bloqade.atom_arrangement import Chain
from bloqade.task.batch import RemoteBatch
import glob
import os
import pytest


def test_batch_error(*args):
    with pytest.raises(RemoteBatch.SubmissionException):
        (
            Chain(5, 6.1)
            .rydberg.detuning.uniform.linear(-10, 10, 3.0)
            .quera.mock(submission_error=True)
            .run_async(100)
        )

    error_files = glob.glob("partial-batch-errors-*")
    batch_files = glob.glob("partial-batch-future-*")

    for error_file, batch_file in zip(error_files, batch_files):
        os.remove(error_file)
        os.remove(batch_file)

    assert len(error_files) == 1
    assert len(batch_files) == 1


def test_batch_warn():
    with pytest.warns():
        (
            Chain(5, 6.1)
            .rydberg.detuning.uniform.linear(-10, 10, 3.0)
            .quera.mock(submission_error=True)
            .run_async(100, ignore_submission_error=True)
        )

    error_files = glob.glob("partial-batch-errors-*")
    batch_files = glob.glob("partial-batch-future-*")

    for error_file, batch_file in zip(error_files, batch_files):
        os.remove(error_file)
        os.remove(batch_file)

    assert len(error_files) == 1
    assert len(batch_files) == 1



import bloqade.ir.location as location
from bloqade.ir import Linear, Constant
from bloqade.serialize import loads, dumps


def test():
    batch = (
        location.Square(6)
        .rydberg.detuning.uniform.apply(
            Constant("initial_detuning", "up_time")
            .append(Linear("initial_detuning", "final_detuning", "anneal_time"))
            .append(Constant("final_detuning", "up_time"))
        )
        .rabi.amplitude.uniform.apply(
            Linear(0.0, "rabi_amplitude_max", "up_time")
            .append(Constant("rabi_amplitude_max", "anneal_time"))
            .append(Linear("rabi_amplitude_max", 0.0, "up_time"))
        )
        .assign(
            initial_detuning=-10,
            up_time=0.1,
            final_detuning=15,
            anneal_time=10,
            rabi_amplitude_max=15,
        )
        .quera.mock()
        ._compile(shots=10)
    )

    batch_str = dumps(batch)
    assert isinstance(loads(batch_str), type(batch))



import bloqade.ir.location as location
import bloqade.ir.control.sequence as sequence
import bloqade.ir.control.pulse as pulse
import bloqade.ir.control.field as field
import bloqade.ir.analog_circuit as analog_circuit
import bloqade.ir.control.waveform as waveform
from bloqade.ir.location import (
    Chain,
    Honeycomb,
    Kagome,
    Lieb,
    ListOfLocations,
    Rectangular,
    Square,
    Triangular,
)
from bloqade.ir.location.base import AtomArrangement, LocationInfo, ParallelRegister
import bloqade.ir.scalar as scalar

from bloqade.ir.visitor.analog_circuit import AnalogCircuitVisitor
from bloqade.ir.visitor.location import LocationVisitor
from bloqade.ir.visitor.waveform import WaveformVisitor
from bloqade.ir.visitor.scalar import ScalarVisitor

import json

from typing import Any, Dict


class ScalarSerilaizer(ScalarVisitor):
    def visit_literal(self, ast: scalar.Literal) -> Dict[str, Dict[str, str]]:
        return {"literal": {"value": str(ast.value)}}

    def visit_variable(self, ast: scalar.Variable) -> Dict[str, Dict[str, str]]:
        return {"variable": {"name": ast.name}}

    def visit_assigned_variable(self, ast: scalar.AssignedVariable) -> Dict[str, Any]:
        return {
            "assigned_variable": {
                "name": ast.name,
                "value": str(ast.value),
            }
        }

    def visit_negative(self, ast: scalar.Negative) -> Dict[str, Any]:
        return {"negative": {"expr": self.visit(ast.expr)}}

    def visit_add(self, ast: scalar.Add) -> Dict[str, Any]:
        return {"add": {"lhs": self.visit(ast.lhs), "rhs": self.visit(ast.rhs)}}

    def visit_mul(self, ast: scalar.Mul) -> Dict[str, Any]:
        return {"mul": {"lhs": self.visit(ast.lhs), "rhs": self.visit(ast.rhs)}}

    def visit_div(self, ast: scalar.Max) -> Dict[str, Any]:
        return {"div": {"lhs": self.visit(ast.lhs), "rhs": self.visit(ast.rhs)}}

    def visit_min(self, ast: scalar.Min) -> Dict[str, Any]:
        return {"min": {"exprs": list(map(self.visit, ast.exprs))}}

    def visit_max(self, ast: scalar.Min) -> Dict[str, Any]:
        return {"max": {"exprs": list(map(self.visit, ast.exprs))}}

    def visit_slice(self, ast: scalar.Slice):
        return {
            "slice_scalar": {
                "expr": self.visit(ast.expr),
                "interval": self.visit(ast.interval),
            }
        }

    def visit_interval(self, ast: scalar.Interval) -> Dict[str, Any]:
        return {
            "interval": {
                "start": self.visit(ast.start) if ast.start is not None else None,
                "stop": self.visit(ast.stop) if ast.stop is not None else None,
            }
        }

    def default(self, obj: Any) -> Dict[str, Any]:
        if isinstance(obj, scalar.Scalar):
            return self.visit(obj)

        return super().default(obj)


class WaveformSerializer(WaveformVisitor):
    def __init__(self):
        self.scalar_encoder = ScalarSerilaizer()

    def visit_constant(self, ast: waveform.Constant) -> Dict[str, Any]:
        return {
            "constant": {
                "value": self.scalar_encoder.visit(ast.value),
                "duration": self.scalar_encoder.visit(ast.duration),
            }
        }

    def visit_linear(self, ast: waveform.Linear) -> Dict[str, Any]:
        return {
            "linear": {
                "start": self.scalar_encoder.visit(ast.start),
                "stop": self.scalar_encoder.visit(ast.stop),
                "duration": self.scalar_encoder.visit(ast.duration),
            }
        }

    def visit_poly(self, ast: waveform.Poly) -> Dict[str, Any]:
        return {
            "poly": {
                "coeffs": list(map(self.scalar_encoder.visit, ast.coeffs)),
                "duration": self.scalar_encoder.visit(ast.duration),
            }
        }

    def visit_python_fn(self, ast: waveform.PythonFn) -> Dict[str, Any]:
        raise ValueError("Bloqade does not support serialization of Python code.")

    def visit_negative(self, ast: waveform.Negative) -> Dict[str, Any]:
        return {"negative_waveform": {"waveform": self.visit(ast.waveform)}}

    def visit_add(self, ast: waveform.Add) -> Dict[str, Any]:
        return {
            "add_waveform": {
                "left": self.visit(ast.left),
                "right": self.visit(ast.right),
            }
        }

    def visit_scale(self, ast: waveform.Scale) -> Dict[str, Any]:
        return {
            "scale": {
                "waveform": self.visit(ast.waveform),
                "scalar": self.scalar_encoder.visit(ast.scalar),
            }
        }

    def visit_slice(self, ast: waveform.Slice) -> Dict[str, Any]:
        return {
            "slice_waveform": {
                "waveform": self.visit(ast.waveform),
                "interval": self.scalar_encoder.visit(ast.interval),
            }
        }

    def visit_sample(self, ast: waveform.Sample) -> Dict[str, Any]:
        return {
            "sample": {
                "waveform": self.visit(ast.waveform),
                "dt": self.scalar_encoder.visit(ast.dt),
                "interpolation": ast.interpolation.value,
            }
        }

    def visit_append(self, ast: waveform.Append) -> Dict[str, Any]:
        return {"append_waveform": {"waveforms": list(map(self.visit, ast.waveforms))}}

    def visit_record(self, ast: waveform.Record) -> Dict[str, Any]:
        return {
            "record": {
                "var": self.scalar_encoder(ast.var),
                "waveform": self.visit(ast.waveform),
            }
        }

    def visit_smooth(self, ast: waveform.Smooth) -> Dict[str, Any]:
        return {
            "smooth": {
                "waveform": self.visit(ast.waveform),
                "radius": self.scalar_encoder.visit(ast.radius),
                "kernel": type(ast.kernel).__name__,
            }
        }

    def visit_alligned(self, ast: waveform.AlignedWaveform) -> Dict[str, Any]:
        if isinstance(ast.value, scalar.Scalar):
            value = self.scalar_encoder.visit(ast.value)
        else:
            value = ast.value.value
        return {
            "alligned": {
                "waveform": self.visit(ast.waveform),
                "allignment": self.scalar_encoder.visit(ast.allignment),
                "value": value,
            }
        }

    def default(self, obj: Any) -> Dict[str, Any]:
        if isinstance(obj, waveform.Waveform):
            return self.visit(obj)

        return super().default(obj)


class LocationSerializer(LocationVisitor):
    def __init__(self):
        self.scalar_serializer = ScalarSerilaizer()

    def visit_location_info(self, info: LocationInfo) -> Any:
        return {
            "location_info": {
                "position": list(map(self.scalar_serializer.visit, info.position)),
                "filled": bool(info.filling.value),
            }
        }

    def visit_chain(self, ast: Chain) -> Any:
        return {
            "chain": {
                "lattice_spacing": self.scalar_serializer.visit(ast.lattice_spacing),
                "L": ast.shape[0],
                "vertical_chain": ast.vertical,
            }
        }

    def visit_honeycomb(self, ast: Honeycomb) -> Any:
        return {
            "honeycomb": {
                "lattice_spacing": self.scalar_serializer.visit(ast.lattice_spacing),
                "L1": ast.shape[0],
                "L2": ast.shape[1],
            }
        }

    def visit_kagome(self, ast: Kagome) -> Any:
        return {
            "kagome": {
                "lattice_spacing": self.scalar_serializer.visit(ast.lattice_spacing),
                "L1": ast.shape[0],
                "L2": ast.shape[1],
            }
        }

    def visit_lieb(self, ast: Lieb) -> Any:
        return {
            "lieb": {
                "lattice_spacing": self.scalar_serializer.visit(ast.lattice_spacing),
                "L1": ast.shape[0],
                "L2": ast.shape[1],
            }
        }

    def visit_list_of_locations(self, ast: ListOfLocations) -> Any:
        return {
            "list_of_locations": {
                "location_list": list(map(self.visit, ast.location_list))
            }
        }

    def visit_rectangular(self, ast: Rectangular) -> Any:
        return {
            "rectangular": {
                "lattice_spacing_x": self.scalar_serializer.visit(
                    ast.lattice_spacing_x
                ),
                "lattice_spacing_y": self.scalar_serializer.visit(
                    ast.lattice_spacing_y
                ),
                "width": ast.shape[0],
                "height": ast.shape[1],
            }
        }

    def visit_square(self, ast: Square) -> Any:
        return {
            "square": {
                "lattice_spacing": self.visit(
                    self.scalar_serializer.visit(ast.lattice_spacing)
                ),
                "L1": ast.shape[0],
                "L2": ast.shape[1],
            }
        }

    def visit_triangular(self, ast: Triangular) -> Any:
        return {
            "triangular": {
                "lattice_spacing": self.scalar_serializer.visit(ast.lattice_spacing),
                "L1": ast.shape[0],
                "L2": ast.shape[1],
            }
        }

    def visit_parallel_register(self, ast: location.ParallelRegister) -> Any:
        return {
            "parallel_register": {
                "register": self.visit(ast._register),
                "cluster_spacing": self.visit(ast._cluster_spacing),
            }
        }


class AnalogCircuitSerializer(AnalogCircuitVisitor):
    def __init__(self) -> None:
        self.waveform_serializer = WaveformSerializer()
        self.scalar_serializer = ScalarSerilaizer()
        self.location_serializer = LocationSerializer()

    def visit_register(self, ast: AtomArrangement) -> Any:
        return self.location_serializer.visit(ast)

    def visit_parallel_register(self, ast: ParallelRegister) -> Any:
        return self.location_serializer.visit(ast)

    def visit_sequence(self, ast: sequence.Sequence) -> Dict[str, Any]:
        object_to_str = {
            sequence.rydberg: "rydberg",
            sequence.hyperfine: "hyperfine",
        }
        return {
            "sequence": {
                "pulses": {
                    object_to_str[k]: self.visit(v) for k, v in ast.pulses.items()
                }
            }
        }

    def visit_named_sequence(self, ast: sequence.NamedSequence) -> Dict[str, Any]:
        return {
            "named_sequence": {
                "name": ast.name,
                "sub_sequence": self.visit(ast.sub_sequence),
            }
        }

    def visit_append_sequence(self, ast: sequence.Append) -> Dict[str, Any]:
        return {
            "append_sequence": {"sequences": [self.visit(s) for s in ast.sequences]}
        }

    def visit_slice_sequence(self, ast: sequence.Slice) -> Dict[str, Any]:
        return {
            "slice_sequence": {
                "sequence": self.visit(ast.sequence),
                "interval": self.scalar_serializer.visit(ast.interval),
            }
        }

    def visit_pulse(self, ast: pulse.Pulse):
        object_map_str = {
            pulse.detuning: "detuning",
            pulse.rabi.amplitude: "rabi_frequency_amplitude",
            pulse.rabi.phase: "rabi_frequency_phase",
        }
        return {
            "pulse": {
                "fields": {
                    object_map_str[k]: self.visit(v) for k, v in ast.fields.items()
                }
            }
        }

    def visit_append_pulse(self, ast: pulse.Append) -> Dict[str, Any]:
        return {"append_pulse": {"pulses": [self.visit(p) for p in ast.pulses]}}

    def visit_named_pulse(self, ast: pulse.NamedPulse) -> Dict[str, Any]:
        return {
            "named_pulse": {"name": ast.name, "sub_pulse": self.visit(ast.sub_pulse)}
        }

    def visit_slice_pulse(self, ast: pulse.Slice) -> Dict[str, Any]:
        return {
            "slice_pulse": {
                "pulse": self.visit(ast.sub_pulse),
                "interval": self.scalar_serializer.visit(ast.interval),
            }
        }

    def visit_location(self, ast: field.Location) -> Dict[str, Any]:
        return {"location": {"label": ast.label}}

    def visit_scaled_locations(self, ast: field.ScaledLocations) -> Dict[str, Any]:
        return {
            "scaled_locations": [(self.visit(k), self.visit(v)) for k, v in ast.items()]
        }

    def visit_uniform(self, ast: field.Uniform) -> Dict[str, Any]:
        return {"uniform": {}}

    def visit_run_time_vector(self, ast: field.RunTimeVector) -> Dict[str, Any]:
        return {"run_time_vector": {"name": ast.name}}

    def visit_assigned_run_time_vector(
        self, ast: field.AssignedRunTimeVector
    ) -> Dict[str, Any]:
        return {
            "assigned_run_time_vector": {
                "name": ast.name,
                "value": [str(v) for v in ast.value],
            }
        }

    def visit_field(self, ast: field.Field) -> Any:
        return {
            "field": {
                "value": [(self.visit(k), self.visit(v)) for k, v in ast.drives.items()]
            }
        }

    def visit_waveform(self, ast: waveform.Waveform) -> Any:
        return self.waveform_serializer.visit(ast)

    def visit_analog_circuit(self, ast: analog_circuit.AnalogCircuit) -> Any:
        return {
            "analog_circuit": {
                "sequence": self.visit(ast.sequence),
                "register": self.visit(ast.register),
                "static_params": ast.static_params,
                "batch_params": ast.batch_params,
                "order": ast.order,
            }
        }


class BloqadeIRSerializer(json.JSONEncoder):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.analog_circuit_serializer = AnalogCircuitSerializer()
        self.waveform_serializer = WaveformSerializer()
        self.scalar_serializer = ScalarSerilaizer()
        self.bloqade_seq_types = (
            analog_circuit.AnalogCircuit,
            location.AtomArrangement,
            location.ParallelRegister,
            analog_circuit.AnalogCircuit,
            sequence.SequenceExpr,
            pulse.PulseExpr,
            field.FieldExpr,
        )

    def default(self, o: Any) -> Any:
        if isinstance(o, self.bloqade_seq_types):
            return self.analog_circuit_serializer.visit(o)
        elif isinstance(o, waveform.Waveform):
            return self.waveform_serializer.visit(o)
        elif isinstance(o, scalar.Scalar):
            return self.scalar_serializer.visit(o)
        else:
            return super().default(o)


class BloqadeIRDeserializer:
    constructors = {
        "literal": scalar.Literal,
        "variable": scalar.Variable,
        "assigned_variable": scalar.AssignedVariable,
        "negative": scalar.Negative,
        "add": scalar.Add,
        "mul": scalar.Mul,
        "div": scalar.Div,
        "min": scalar.Min,
        "max": scalar.Max,
        "slice_scalar": scalar.Slice,
        "interval": scalar.Interval,
        "constant": waveform.Constant,
        "linear": waveform.Linear,
        "poly": waveform.Poly,
        "negative_waveform": waveform.Negative,
        "add_waveform": waveform.Add,
        "scale": waveform.Scale,
        "slice_waveform": waveform.Slice,
        "sample": waveform.Sample,
        "append_waveform": waveform.Append,
        "record": waveform.Record,
        "smooth": waveform.Smooth,
        "alligned": waveform.AlignedWaveform,
        "chain": location.Chain,
        "honeycomb": location.Honeycomb,
        "kagome": location.Kagome,
        "lieb": location.Lieb,
        "list_of_locations": location.ListOfLocations,
        "rectangular": location.Rectangular,
        "square": location.Square,
        "triangular": location.Triangular,
        "parallel_register": location.ParallelRegister,
        "sequence": sequence.Sequence,
        "named_sequence": sequence.NamedSequence,
        "append_sequence": sequence.Append,
        "slice_sequence": sequence.Slice,
        "pulse": pulse.Pulse,
        "append_pulse": pulse.Append,
        "named_pulse": pulse.NamedPulse,
        "slice_pulse": pulse.Slice,
        "location": field.Location,
        "scaled_locations": field.ScaledLocations,
        "uniform": field.Uniform,
        "run_time_vector": field.RunTimeVector,
        "assigned_run_time_vector": field.AssignedRunTimeVector,
        "field": field.Field,
        "analog_circuit": analog_circuit.AnalogCircuit,
    }

    @classmethod
    def object_hook(cls, obj):
        if isinstance(obj, dict) and len(obj) == 1:
            ((head, options),) = obj.items()
            if head in cls.constructors:
                return cls.constructors[head](**options)

        return obj



from functools import cached_property
from bloqade.ir.analog_circuit import AnalogCircuit
from bloqade.ir.scalar import Literal
import bloqade.ir.control.field as field
import bloqade.ir.control.pulse as pulse
import bloqade.ir.control.sequence as sequence

from bloqade.ir.location.base import AtomArrangement, ParallelRegister
from bloqade.ir.visitor.analog_circuit import AnalogCircuitVisitor

import bloqade.submission.ir.task_specification as task_spec
from bloqade.submission.ir.braket import BraketTaskSpecification
from bloqade.submission.ir.task_specification import QuEraTaskSpecification
from bloqade.submission.ir.parallel import ParallelDecoder, ClusterLocationInfo
from bloqade.submission.ir.capabilities import QuEraCapabilities
from bloqade.codegen.hardware.piecewise_linear import (
    PiecewiseLinearCodeGen,
    PiecewiseLinear,
)
from bloqade.codegen.hardware.piecewise_constant import (
    PiecewiseConstantCodeGen,
    PiecewiseConstant,
)

from typing import Any, Dict, Tuple, List, Union, Optional
from pydantic.dataclasses import dataclass
import numbers
from decimal import Decimal
import numpy as np


@dataclass(frozen=True)
class AHSCodegenResult:
    nshots: int
    parallel_decoder: Optional[ParallelDecoder]
    sites: List[Tuple[Decimal, Decimal]]
    filling: List[bool]
    global_detuning: PiecewiseLinear
    global_rabi_amplitude: PiecewiseLinear
    global_rabi_phase: PiecewiseConstant
    lattice_site_coefficients: Optional[List[Decimal]]
    local_detuning: Optional[PiecewiseLinear]

    def slice(self, start_time: Decimal, stop_time: Decimal) -> "AHSCodegenResult":
        return AHSCodegenResult(
            self.nshots,
            self.parallel_decoder,
            self.sites,
            self.filling,
            self.global_detuning.slice(start_time, stop_time),
            self.global_rabi_amplitude.slice(start_time, stop_time),
            self.global_rabi_phase.slice(start_time, stop_time),
            self.lattice_site_coefficients,
            self.local_detuning.slice(start_time, stop_time),
        )

    def append(self, other: "AHSCodegenResult") -> "AHSCodegenResult":
        assert self.nshots == other.nshots
        assert self.parallel_decoder == other.parallel_decoder
        assert self.sites == other.sites
        assert self.filling == other.filling
        assert self.lattice_site_coefficients == other.lattice_site_coefficients

        return AHSCodegenResult(
            self.nshots,
            self.parallel_decoder,
            self.sites,
            self.filling,
            self.global_detuning.append(other.global_detuning),
            self.global_rabi_amplitude.append(other.global_rabi_amplitude),
            self.global_rabi_phase.append(other.global_rabi_phase),
            self.lattice_site_coefficients,
            self.local_detuning.append(other.local_detuning),
        )

    @staticmethod
    def convert_position_units(position):
        return tuple(coordinate * Decimal("1e-6") for coordinate in position)

    @staticmethod
    def convert_time_units(time):
        return Decimal("1e-6") * time

    @staticmethod
    def convert_energy_units(energy):
        return Decimal("1e6") * energy

    @cached_property
    def braket_task_ir(self) -> BraketTaskSpecification:
        import braket.ir.ahs as ir

        return BraketTaskSpecification(
            nshots=self.nshots,
            program=ir.Program(
                setup=ir.Setup(
                    ahs_register=ir.AtomArrangement(
                        sites=list(map(self.convert_position_units, self.sites)),
                        filling=self.filling,
                    )
                ),
                hamiltonian=ir.Hamiltonian(
                    drivingFields=[
                        ir.DrivingField(
                            amplitude=ir.PhysicalField(
                                time_series=ir.TimeSeries(
                                    times=list(
                                        map(
                                            self.convert_time_units,
                                            self.global_rabi_amplitude.times,
                                        )
                                    ),
                                    values=list(
                                        map(
                                            self.convert_energy_units,
                                            self.global_rabi_amplitude.values,
                                        )
                                    ),
                                ),
                                pattern="uniform",
                            ),
                            phase=ir.PhysicalField(
                                time_series=ir.TimeSeries(
                                    times=list(
                                        map(
                                            self.convert_time_units,
                                            self.global_rabi_phase.times,
                                        )
                                    ),
                                    values=self.global_rabi_phase.values,
                                ),
                                pattern="uniform",
                            ),
                            detuning=ir.PhysicalField(
                                time_series=ir.TimeSeries(
                                    times=list(
                                        map(
                                            self.convert_time_units,
                                            self.global_detuning.times,
                                        )
                                    ),
                                    values=list(
                                        map(
                                            self.convert_energy_units,
                                            self.global_detuning.values,
                                        )
                                    ),
                                ),
                                pattern="uniform",
                            ),
                        )
                    ],
                    shiftingFields=(
                        []
                        if self.lattice_site_coefficients is None
                        else [
                            ir.ShiftingField(
                                magnitude=ir.PhysicalField(
                                    time_series=ir.TimeSeries(
                                        times=list(
                                            map(
                                                self.convert_time_units,
                                                self.local_detuning.times,
                                            )
                                        ),
                                        values=list(
                                            map(
                                                self.convert_energy_units,
                                                self.local_detuning.values,
                                            )
                                        ),
                                    ),
                                    pattern=self.lattice_site_coefficients,
                                )
                            )
                        ]
                    ),
                ),
            ),
        )

    @cached_property
    def quera_task_ir(self) -> QuEraTaskSpecification:
        import bloqade.submission.ir.task_specification as task_spec

        return task_spec.QuEraTaskSpecification(
            nshots=self.nshots,
            lattice=task_spec.Lattice(
                sites=list(map(self.convert_position_units, self.sites)),
                filling=self.filling,
            ),
            effective_hamiltonian=task_spec.EffectiveHamiltonian(
                rydberg=task_spec.RydbergHamiltonian(
                    rabi_frequency_amplitude=task_spec.RabiFrequencyAmplitude(
                        global_=task_spec.GlobalField(
                            times=list(
                                map(
                                    self.convert_time_units,
                                    self.global_rabi_amplitude.times,
                                )
                            ),
                            values=list(
                                map(
                                    self.convert_energy_units,
                                    self.global_rabi_amplitude.values,
                                )
                            ),
                        )
                    ),
                    rabi_frequency_phase=task_spec.RabiFrequencyPhase(
                        global_=task_spec.GlobalField(
                            times=list(
                                map(
                                    self.convert_time_units,
                                    self.global_rabi_phase.times,
                                )
                            ),
                            values=self.global_rabi_phase.values,
                        )
                    ),
                    detuning=task_spec.Detuning(
                        global_=task_spec.GlobalField(
                            times=list(
                                map(self.convert_time_units, self.global_detuning.times)
                            ),
                            values=list(
                                map(
                                    self.convert_energy_units,
                                    self.global_detuning.values,
                                )
                            ),
                        ),
                        local=(
                            None
                            if self.lattice_site_coefficients is None
                            else task_spec.LocalField(
                                times=list(
                                    map(
                                        self.convert_time_units,
                                        self.local_detuning.times,
                                    )
                                ),
                                values=list(
                                    map(
                                        self.convert_energy_units,
                                        self.local_detuning.values,
                                    )
                                ),
                                lattice_site_coefficients=self.lattice_site_coefficients,
                            )
                        ),
                    ),
                )
            ),
        )


class AHSCodegen(AnalogCircuitVisitor):
    def __init__(
        self,
        shots: int,
        assignments: Dict[str, Union[numbers.Real, List[numbers.Real]]] = {},
        capabilities: Optional[QuEraCapabilities] = None,
    ):
        self.nshots = shots
        self.capabilities = capabilities
        self.assignments = assignments
        self.parallel_decoder = None
        self.sites = []
        self.filling = []
        self.global_detuning = None
        self.local_detuning = None
        self.lattice_site_coefficients = None
        self.global_rabi_amplitude = None
        self.global_rabi_phase = None

    def extract_fields(self, ahs_result: AHSCodegenResult) -> None:
        self.nshots = ahs_result.nshots
        self.sites = ahs_result.sites
        self.filling = ahs_result.filling
        self.global_detuning = ahs_result.global_detuning
        self.global_rabi_amplitude = ahs_result.global_rabi_amplitude
        self.global_rabi_phase = ahs_result.global_rabi_phase
        self.lattice_site_coefficients = ahs_result.lattice_site_coefficients
        self.local_detuning = ahs_result.local_detuning

    def fix_up_missing_fields(self) -> None:
        # fix-up any missing fields
        duration = 0.0

        if self.global_rabi_amplitude:
            duration = max(duration, self.global_rabi_amplitude.times[-1])

        if self.global_rabi_phase:
            duration = max(duration, self.global_rabi_phase.times[-1])

        if self.global_detuning:
            duration = max(duration, self.global_detuning.times[-1])

        if self.local_detuning:
            duration = max(duration, self.local_detuning.times[-1])

        if duration > 0:
            if self.global_rabi_amplitude is None:
                self.global_rabi_amplitude = PiecewiseLinear(
                    [Decimal(0), duration], [Decimal(0), Decimal(0)]
                )

            if self.global_rabi_phase is None:
                self.global_rabi_phase = PiecewiseConstant(
                    [Decimal(0), duration], [Decimal(0), Decimal(0)]
                )

            if self.global_detuning is None:
                self.global_detuning = PiecewiseLinear(
                    [Decimal(0), duration], [Decimal(0), Decimal(0)]
                )

        if self.local_detuning is None:
            pass

    def post_visit_spatial_modulation(self, lattice_site_coefficients):
        self.lattice_site_coefficients = []
        if self.parallel_decoder:
            for cluster_site_info in self.parallel_decoder.mapping:
                self.lattice_site_coefficients.append(
                    lattice_site_coefficients[cluster_site_info.cluster_location_index]
                )
        else:
            self.lattice_site_coefficients = lattice_site_coefficients

    def visit_location(self, ast: field.Location) -> Any:
        if (
            ast.value >= self.n_atoms
        ):  # n_atoms is now the number of atoms in the parallelized
            # lattice, but needs to be num. atoms in the original
            raise ValueError(
                f"field.Location({ast.value}) is larger than the register."
            )

    def visit_uniform_modulation(self, ast: field.UniformModulation) -> Any:
        lattice_site_coefficients = [Decimal("1.0") for _ in range(self.n_atoms)]
        self.post_visit_spatial_modulation(lattice_site_coefficients)

    def visit_scaled_locations(self, ast: field.ScaledLocations) -> Any:
        lattice_site_coefficients = []

        for location in ast.value.keys():
            self.visit(location)

        for atom_index in range(self.n_atoms):
            scale = ast.value.get(field.Location(atom_index), Literal(0.0))
            lattice_site_coefficients.append(scale(**self.assignments))

        self.post_visit_spatial_modulation(lattice_site_coefficients)

    def visit_run_time_vector(self, ast: field.RunTimeVector) -> Any:
        if len(self.assignments[ast.name]) != self.n_atoms:
            raise ValueError(
                f"Coefficient list {ast.name} doesn't match the size of register "
                f"{self.n_atoms}."
            )
        lattice_site_coefficients = list(self.assignments[ast.name])
        self.post_visit_spatial_modulation(lattice_site_coefficients)

    def visit_assigned_run_time_vector(self, ast: field.AssignedRunTimeVector) -> Any:
        if len(ast.value) != self.n_atoms:
            raise ValueError(
                f"Coefficient list {ast.name} doesn't match the size of register "
                f"{self.n_atoms}."
            )
        lattice_site_coefficients = ast.value
        self.post_visit_spatial_modulation(lattice_site_coefficients)

    def calculate_detuning(self, ast: field.Field) -> Any:
        if len(ast.drives) == 1 and field.Uniform in ast.drives:
            self.global_detuning = PiecewiseLinearCodeGen(self.assignments).emit(
                ast.drives[field.Uniform]
            )

        elif len(ast.drives) == 1:
            ((spatial_modulation, waveform),) = ast.drives.items()

            self.local_detuning = PiecewiseLinearCodeGen(self.assignments).emit(
                waveform
            )

            self.global_detuning = PiecewiseLinear(
                [Decimal(0), self.local_detuning.times[-1]], [Decimal(0), Decimal(0)]
            )

            self.visit(spatial_modulation)

        elif len(ast.drives) == 2 and field.Uniform in ast.drives:
            # will only be two keys
            for k in ast.drives.keys():
                if k == field.Uniform:
                    self.global_detuning = PiecewiseLinearCodeGen(
                        self.assignments
                    ).emit(ast.drives[field.Uniform])
                else:  # can be field.RunTimeVector or field.ScaledLocations
                    spatial_modulation = k
                    self.local_detuning = PiecewiseLinearCodeGen(self.assignments).emit(
                        ast.drives[k]
                    )

            self.visit(spatial_modulation)  # just visit the non-uniform locations

        else:
            raise ValueError(
                "Failed to compile Detuning to QuEra task, "
                "found more than one non-uniform modulation: "
                f"{repr(ast)}."
            )

    def calculate_rabi_amplitude(self, ast: field.Field) -> Any:
        if len(ast.drives) == 1 and field.Uniform in ast.drives:
            self.global_rabi_amplitude = PiecewiseLinearCodeGen(self.assignments).emit(
                ast.drives[field.Uniform]
            )

        else:
            raise ValueError(
                "Failed to compile Rabi Amplitude to QuEra task, "
                "found non-uniform modulation: "
                f"{repr(ast)}."
            )

    def calculate_rabi_phase(self, ast: field.Field) -> Any:
        if len(ast.drives) == 1 and field.Uniform in ast.drives:  # has to be global
            self.global_rabi_phase = PiecewiseConstantCodeGen(self.assignments).emit(
                ast.drives[field.Uniform]
            )

        else:
            raise ValueError(
                "Failed to compile Rabi Phase to QuEra task, "
                "found non-uniform modulation: "
                f"{repr(ast)}."
            )

    def visit_field(self, ast: field.Field):
        if self.field_name == pulse.detuning:
            self.calculate_detuning(ast)
        elif self.field_name == pulse.rabi.amplitude:
            self.calculate_rabi_amplitude(ast)
        elif self.field_name == pulse.rabi.phase:
            self.calculate_rabi_phase(ast)

    def visit_pulse(self, ast: pulse.Pulse):
        for field_name in ast.fields.keys():
            self.field_name = field_name
            self.visit(ast.fields[field_name])

    def visit_named_pulse(self, ast: pulse.NamedPulse):
        self.visit(ast.pulse)

    def visit_append_pulse(self, ast: pulse.Append):
        subexpr_compiler = AHSCodegen(self.nshots, self.assignments)
        ahs_result = subexpr_compiler.emit(ast.pulses)

        for seq in ast.sequences:
            new_ahs_result = subexpr_compiler.emit(seq)
            ahs_result = ahs_result.append(new_ahs_result)

        self.extract_fields(ahs_result)

    def visit_slice_pulse(self, ast: pulse.Slice):
        start_time = ast.interval.start(**self.assignments)
        stop_time = ast.interval.stop(**self.assignments)

        ahs_result = AHSCodegen(self.nshots, self.assignments).emit(ast.pulse)
        ahs_result = ahs_result.slice(start_time, stop_time)

        self.extract_fields(ahs_result)

    def visit_sequence(self, ast: sequence.Sequence):
        if sequence.HyperfineLevelCoupling() in ast.pulses:
            raise ValueError("QuEra tasks does not support Hyperfine coupling.")

        self.visit(ast.pulses.get(sequence.RydbergLevelCoupling(), pulse.Pulse({})))

    def visit_named_sequence(self, ast: sequence.NamedSequence):
        self.visit(ast.sequence)

    def visit_append_sequence(self, ast: sequence.Append):
        subexpr_compiler = AHSCodegen(self.nshots, self.assignments)
        ahs_result = subexpr_compiler.emit(ast.sequences[0])

        for sub_sequence in ast.sequences[1:]:
            new_ahs_result = subexpr_compiler.emit(sub_sequence)
            ahs_result = ahs_result.append(new_ahs_result)

        self.extract_fields(ahs_result)

    def visit_slice_sequence(self, ast: sequence.Slice):
        start_time = ast.interval.start(**self.assignments)
        stop_time = ast.interval.stop(**self.assignments)

        ahs_result = AHSCodegen(self.nshots, self.assignments).emit(ast.sequence)
        ahs_result = ahs_result.slice(start_time, stop_time)
        self.extract_fields(ahs_result)

    def visit_register(self, ast: AtomArrangement):
        self.sites = []
        self.filling = []

        for location_info in ast.enumerate():
            site = tuple(ele(**self.assignments) for ele in location_info.position)
            self.sites.append(site)
            self.filling.append(location_info.filling.value)

        self.n_atoms = len(self.sites)

    def visit_parallel_register(self, ast: ParallelRegister) -> Any:
        if self.capabilities is None:
            raise NotImplementedError(
                "Cannot parallelize register without device capabilities."
            )

        height_max = self.capabilities.capabilities.lattice.area.height / Decimal(
            "1e-6"
        )
        width_max = self.capabilities.capabilities.lattice.area.width / Decimal("1e-6")
        number_sites_max = (
            self.capabilities.capabilities.lattice.geometry.number_sites_max
        )

        register_filling = np.asarray(ast.register_filling)

        register_locations = np.asarray(
            [
                [s(**self.assignments) for s in location]
                for location in ast.register_locations
            ]
        )
        register_locations = register_locations - register_locations.min(axis=0)

        shift_vectors = np.asarray(
            [
                [s(**self.assignments) for s in shift_vector]
                for shift_vector in ast.shift_vectors
            ]
        )

        # build register by stack method because
        # shift_vectors might not be rectangular
        c_stack = [(0, 0)]
        visited = set([(0, 0)])
        mapping = []
        global_site_index = 0
        sites = []
        filling = []
        while c_stack:
            if len(mapping) + len(ast.register_locations) > number_sites_max:
                break

            cluster_index = c_stack.pop()

            shift = (
                shift_vectors[0] * cluster_index[0]
                + shift_vectors[1] * cluster_index[1]
            )

            new_register_locations = register_locations + shift

            # skip clusters that fall out of bounds
            if (
                np.any(new_register_locations < 0)
                or np.any(new_register_locations[:, 0] > width_max)
                or np.any(new_register_locations[:, 1] > height_max)
            ):
                continue

            new_cluster_indices = [
                (cluster_index[0] + 1, cluster_index[1]),
                (cluster_index[0], cluster_index[1] + 1),
                (cluster_index[0] - 1, cluster_index[1]),
                (cluster_index[0], cluster_index[1] - 1),
            ]

            for new_cluster_index in new_cluster_indices:
                if new_cluster_index not in visited:
                    visited.add(new_cluster_index)
                    c_stack.append(new_cluster_index)

            for cluster_location_index, (location, filled) in enumerate(
                zip(new_register_locations[:], register_filling)
            ):
                site = tuple(location)
                sites.append(site)
                filling.append(filled)

                mapping.append(
                    ClusterLocationInfo(
                        cluster_index=cluster_index,
                        global_location_index=global_site_index,
                        cluster_location_index=cluster_location_index,
                    )
                )

                global_site_index += 1

        self.lattice = task_spec.Lattice(sites=sites, filling=filling)
        self.n_atoms = len(ast.register_locations)
        self.parallel_decoder = ParallelDecoder(mapping=mapping)

    def visit_analog_circuit(self, ast: AnalogCircuit) -> Any:
        self.visit(ast.register)
        self.visit(ast.sequence)

    def emit(self, ast) -> AHSCodegenResult:
        self.visit(ast)
        self.fix_up_missing_fields()

        if all(  # TODO: move this into analysis portion.
            [
                self.global_detuning is None,
                self.global_rabi_amplitude is None,
                self.global_rabi_phase is None,
                self.local_detuning is None,
            ]
        ):
            raise ValueError("No fields were specified.")

        return AHSCodegenResult(
            nshots=self.nshots,
            parallel_decoder=self.parallel_decoder,
            sites=self.sites,
            filling=self.filling,
            global_detuning=self.global_detuning,
            global_rabi_amplitude=self.global_rabi_amplitude,
            global_rabi_phase=self.global_rabi_phase,
            lattice_site_coefficients=self.lattice_site_coefficients,
            local_detuning=self.local_detuning,
        )



import bloqade.ir as ir
from bloqade.ir.location import ListOfLocations, AtomArrangement
from bloqade.ir.location import Square
from bloqade.constants import RB_C6
from bloqade import cast
import pytest
import numpy as np


def test_rydberg_interactions():
    geometry = ListOfLocations([(0, 0), (1, 0), (0, 1), (1, 1)]).scale(5.0)

    V_ij = geometry.rydberg_interaction()

    # 2  3
    #  \/
    #  /\
    # 0  1

    d_01 = 5.0
    d_02 = 5.0
    d_03 = np.sqrt(2) * 5.0
    d_12 = np.sqrt(2) * 5.0
    d_13 = 5.0
    d_23 = 5.0

    V_01 = RB_C6 / d_01**6
    V_02 = RB_C6 / d_02**6
    V_03 = RB_C6 / d_03**6
    V_12 = RB_C6 / d_12**6
    V_13 = RB_C6 / d_13**6
    V_23 = RB_C6 / d_23**6

    V_ij_expected = np.array(
        [[0, 0, 0, 0], [V_01, 0, 0, 0], [V_02, V_12, 0, 0], [V_03, V_13, V_23, 0]]
    )

    assert np.allclose(V_ij, V_ij_expected)


def test_listOfLocations():
    lattice = ListOfLocations(
        [(0, 0), (0, 1), (1, 0), (1, 1)]
    )  # enumerate method gives a generator that can be taken advantage of

    positions1 = set(map(lambda info: info.position, lattice.enumerate()))
    positions_expected = set(
        cast(
            [
                (0, 0),
                (1, 0),
                (0, 1),
                (1, 1),
            ]
        )
    )
    assert positions1 == positions_expected

    # scale:
    lattice = lattice.scale(3.2)

    positions1 = set(map(lambda info: info.position, lattice.enumerate()))
    positions_expected = set(
        cast(
            [
                (0, 0),
                (3.2, 0),
                (0, 3.2),
                (3.2, 3.2),
            ]
        )
    )
    assert positions1 == positions_expected


def test_addlocs_on_square():
    lattice = Square(2, lattice_spacing=1.0)
    lattice1 = lattice.add_position([(9, 6), (4, 4)])

    # type(lattice)
    assert type(lattice1) == ir.location.ListOfLocations

    positions1 = set(map(lambda info: info.position, lattice1.enumerate()))
    positions_expected1 = set(cast([(0, 0), (1, 0), (0, 1), (1, 1), (9, 6), (4, 4)]))

    assert positions1 == positions_expected1


def test_addlocs_filling_options():
    lattice = Square(2, lattice_spacing=1.0)
    lattice = lattice.add_position([(9, 6), (4, 4)], filling=[False, True])

    # type(lattice)
    assert type(lattice) == ir.location.ListOfLocations

    for info in lattice.enumerate():
        if info.position == cast((9, 6)):
            assert bool(info.filling) is False
        else:
            assert bool(info.filling) is True

    assert lattice.n_dims == 2


def test_internal_base_listofloc():
    lattice = AtomArrangement()

    with pytest.raises(NotImplementedError):
        set(map(lambda info: info.position, lattice.enumerate()))

    with pytest.raises(NotImplementedError):
        lattice.n_atoms

    with pytest.raises(NotImplementedError):
        lattice.n_dims



from bokeh.models import Div
from bokeh.layouts import row, column
from bokeh.io import show
from bloqade.visualization import report_visualize
from bokeh.models import (
    NumericInput,
    Button,
    CustomJS,
)


## unify the entry point:
def display_ir(obj, assignments):
    from bloqade.ir.analog_circuit import AnalogCircuit
    from bloqade.ir.control.pulse import PulseExpr
    from bloqade.ir.control.sequence import SequenceExpr
    from bloqade.ir.control.field import Field, SpatialModulation
    from bloqade.ir.control.waveform import Waveform
    from bloqade.ir.location import AtomArrangement

    if isinstance(obj, AnalogCircuit):
        display_analog_circuit(obj, assignments)
    elif isinstance(obj, PulseExpr):
        display_pulse(obj, assignments)
    elif isinstance(obj, SequenceExpr):
        display_sequence(obj, assignments)
    elif isinstance(obj, Field):
        display_field(obj, assignments)
    elif isinstance(obj, SpatialModulation):
        display_spatialmod(obj, assignments)
    elif isinstance(obj, Waveform):
        display_waveform(obj, assignments)
    elif isinstance(obj, AtomArrangement):
        display_atom_arrangement(obj, assignments)
    else:
        raise NotImplementedError(f"not supported IR for display, got {type(obj)}")


def figure_ir(obj, assignments):
    from bloqade.ir.analog_circuit import AnalogCircuit
    from bloqade.ir.control.pulse import PulseExpr
    from bloqade.ir.control.sequence import SequenceExpr
    from bloqade.ir.control.field import Field, SpatialModulation
    from bloqade.ir.control.waveform import Waveform
    from bloqade.ir.location import AtomArrangement

    if isinstance(obj, AnalogCircuit):
        return analog_circuit_figure(obj, assignments)
    elif isinstance(obj, PulseExpr):
        return obj.figure(**assignments)
    elif isinstance(obj, SequenceExpr):
        return obj.figure(**assignments)
    elif isinstance(obj, Field):
        return obj.figure(**assignments)
    elif isinstance(obj, SpatialModulation):
        return obj.figure(**assignments)
    elif isinstance(obj, Waveform):
        return obj.figure(**assignments)
    elif isinstance(obj, AtomArrangement):
        return atom_arrangement_figure(obj, assignments)
    else:
        raise NotImplementedError(f"not supported IR for display, got {type(obj)}")


def liner(txt):
    return f"<p>{txt}</p>"


def builder_figure(builder, batch_id):
    from bloqade.builder.parse.builder import Parser

    routine = Parser().parse(builder)

    analog_circ = routine.circuit
    metas = routine.params

    kwargs = metas.static_params
    kwargs.update(metas.batch_params[batch_id])

    out = "<p>Assignments: </p>"
    for key, val in kwargs.items():
        out += liner(f" :: {key} = {val}")

    field, reg = analog_circ.figure(**kwargs)

    div = Div(text=out, width=200, height=200)
    return row(field, column(reg, div))


def display_builder(builder, batch_id):
    fig = builder_figure(builder, batch_id)
    show(fig)


## ir
def analog_circuit_figure(analog_circuit, assignments):
    return row(*analog_circuit.figure(**assignments))


def display_analog_circuit(analog_circuit, assignments):
    fig = analog_circuit_figure(analog_circuit, assignments)
    show(fig)


def display_pulse(pulse, assignments):
    show(pulse.figure(**assignments))


def display_sequence(sequence, assignments):
    show(sequence.figure(**assignments))


def display_field(field, assignments):
    show(field.figure(**assignments))


def display_spatialmod(spmod, assignments):
    show(spmod.figure(**assignments))


def display_waveform(wvfm_ir, assignments):
    show(wvfm_ir.figure(**assignments))


def atom_arrangement_figure(atom_arrangement, assignments):
    """show the register."""
    p = atom_arrangement.figure(None, **assignments)

    # get the Blocade rad object
    cr = None
    for rd in p.renderers:
        if rd.name == "Brad":
            cr = rd

    # adding rydberg radis input
    Brad_input = NumericInput(
        value=0, low=0, title="Blockade radius (um):", mode="float"
    )

    # js link toggle btn
    toggle_button = Button(label="Toggle")
    toggle_button.js_on_event(
        "button_click",
        CustomJS(args=dict(cr=cr), code="""cr.visible = !cr.visible;"""),
    )

    # js link radius
    Brad_input.js_link("value", cr.glyph, "radius")

    full = column(p, row(Brad_input, toggle_button))
    # full.sizing_mode="scale_both"
    return full


def display_atom_arrangement(atom_arrangement, assignments):
    fig = atom_arrangement_figure(atom_arrangement, assignments)
    show(fig)


## report


def report_figure(report):
    dat = report_visualize.format_report_data(report)
    p = report_visualize.report_visual(*dat)
    return p


def display_report(report):
    fig = report_figure(report)
    show(fig)


## task_ir
def display_task_ir(task_ir):
    show(task_ir.figure())



from IPython.lib.pretty import PrettyPrinter as PP
from bloqade.ir import Linear, Poly
from io import StringIO
from bloqade.ir.tree_print import Printer


def test_printer():
    wf = Linear(start=0, stop=1, duration=1)

    mystdout = StringIO()
    p = PP(mystdout)

    wf._repr_pretty_(p, 1)

    assert (
        mystdout.getvalue()
        == "Linear\n"
        + "├─ start\n"
        + "│  ⇒ Literal: 0\n"
        + "├─ stop\n"
        + "│  ⇒ Literal: 1\n"
        + "└─ duration\n"
        + "   ⇒ Literal: 1"
    )


def test_printer_nodes_list():
    wf = Linear(start=0, stop=1, duration=1)
    wf2 = wf + (wf + wf)

    mystdout = StringIO()
    p = PP(mystdout)

    pr = Printer(p)
    children = wf2.children().copy()  # children is list so no print children
    assert pr.should_print_annotation(children) is False


def test_printer_nodes_dict():
    wf = Poly([1, 2, 3], duration=1)

    mystdout = StringIO()
    p = PP(mystdout)

    pr = Printer(p)
    children = wf.children().copy()  # children is list so no print children

    assert pr.should_print_annotation(children) is True
    wf._repr_pretty_(p, 10)

    assert (
        mystdout.getvalue()
        == "Poly\n"
        + "├─ b\n"
        + "│  ⇒ Literal: 1\n"
        + "├─ t\n"
        + "│  ⇒ Literal: 2\n"
        + "├─ t^2\n"
        + "│  ⇒ Literal: 3\n"
        + "└─ duration\n"
        + "   ⇒ Literal: 1"
    )


def test_printer_nodes_compose_turncate():
    wf = Poly([1, 2, 3], duration=1)
    wf1 = Linear(start=0, stop=1, duration=1)
    wf2 = wf + (wf[:0.5] + wf1)

    mystdout = StringIO()
    p = PP(mystdout)

    wf2._repr_pretty_(p, 0)

    assert mystdout.getvalue() == "+\n├─ Poly\n⋮\n└─ +\n⋮\n"


def test_printer_nodes_compose_all():
    wf = Poly([1, 2, 3], duration=1)
    wf1 = Linear(start=0, stop=1, duration=1)
    wf2 = wf + (wf[:0.5] + wf1)

    mystdout = StringIO()
    p = PP(mystdout)

    wf2._repr_pretty_(p, 10)

    assert (
        mystdout.getvalue()
        == "+\n"
        + "├─ Poly\n"
        + "│  ├─ b\n"
        + "│  │  ⇒ Literal: 1\n"
        + "│  ├─ t\n"
        + "│  │  ⇒ Literal: 2\n"
        + "│  ├─ t^2\n"
        + "│  │  ⇒ Literal: 3\n"
        + "│  └─ duration\n"
        + "│     ⇒ Literal: 1\n"
        + "└─ +\n"
        + "   ├─ Slice\n"
        + "   │  ├─ Poly\n"
        + "   │  │  ├─ b\n"
        + "   │  │  │  ⇒ Literal: 1\n"
        + "   │  │  ├─ t\n"
        + "   │  │  │  ⇒ Literal: 2\n"
        + "   │  │  ├─ t^2\n"
        + "   │  │  │  ⇒ Literal: 3\n"
        + "   │  │  └─ duration\n"
        + "   │  │     ⇒ Literal: 1\n"
        + "   │  └─ Interval\n"
        + "   │     └─ stop\n"
        + "   │        ⇒ Literal: 0.5\n"
        + "   └─ Linear\n"
        + "      ├─ start\n"
        + "      │  ⇒ Literal: 0\n"
        + "      ├─ stop\n"
        + "      │  ⇒ Literal: 1\n"
        + "      └─ duration\n"
        + "         ⇒ Literal: 1"
    )



from bloqade.ir.control.sequence import Sequence
from bloqade.builder.route import PragmaRoute
from bloqade.builder.base import Builder


class SequenceBuilder(PragmaRoute, Builder):
    def __init__(self, sequence: Sequence, parent: Builder):
        super().__init__(parent)
        self._sequence = sequence



from bloqade.visualization.report_visualize import mock_data, report_visual
from bokeh.io import save


def test_report_vis_mock():
    dat = mock_data()

    fig = report_visual(*dat)

    save(fig)
    # from bokeh.models import SVGIcon

    # p = figure(width=200, height=100, toolbar_location=None)
    # p.image_url(url="file:///./logo.png")
    # button = Button(label="", icon=SVGIcon(svg=bloqadeICON(), size=50))
    # show(button)



from .scalar import var, cast, Scalar, Interval, Variable, Literal
from .control.waveform import (
    Waveform,
    Linear,
    Constant,
    Poly,
    Record,
    AlignedWaveform,
    AlignedValue,
    Alignment,
    Interpolation,
    Sample,
    PythonFn,
    to_waveform,
    GaussianKernel,
    LogisticKernel,
    SigmoidKernel,
    TriangleKernel,
    UniformKernel,
    ParabolicKernel,
    BiweightKernel,
    TriweightKernel,
    TricubeKernel,
    CosineKernel,
)
from .control.field import (
    Field,
    Location,
    ScaledLocations,
    Uniform,
    SpatialModulation,
    RunTimeVector,
    AssignedRunTimeVector,
)
from .control.pulse import Pulse, NamedPulse, FieldName, rabi, detuning
from .control.sequence import rydberg, hyperfine, Sequence, LevelCoupling
from .analog_circuit import AnalogCircuit
from .location import (
    AtomArrangement,
    Chain,
    Square,
    Rectangular,
    Honeycomb,
    Triangular,
    Lieb,
    Kagome,
    BoundedBravais,
    ListOfLocations,
    ParallelRegister,
    start,
)

__all__ = [
    "var",
    "cast",
    "Scalar",
    "Interval",
    "Variable",
    "Literal",
    "Linear",
    "Constant",
    "Poly",
    "Record",
    "AlignedWaveform",
    "Alignment",
    "AlignedValue",
    "Waveform",
    "Sample",
    "Interpolation",
    "PythonFn",
    "to_waveform",
    "GaussianKernel",
    "LogisticKernel",
    "SigmoidKernel",
    "TriangleKernel",
    "UniformKernel",
    "ParabolicKernel",
    "BiweightKernel",
    "TriweightKernel",
    "TricubeKernel",
    "CosineKernel",
    "Field",
    "Location",
    "ScaledLocations",
    "Uniform",
    "RunTimeVector",
    "AssignedRunTimeVector",
    "SpatialModulation",
    "Pulse",
    "NamedPulse",
    "FieldName",
    "rabi",
    "detuning",
    "LevelCoupling",
    "rydberg",
    "hyperfine",
    "Sequence",
    "AnalogCircuit",
    ### location ir ###
    "start",
    "AtomArrangement",
    "Chain",
    "Square",
    "Rectangular",
    "Honeycomb",
    "Triangular",
    "Lieb",
    "Kagome",
    "BoundedBravais",
    "ListOfLocations",
    "ParallelRegister",
]



import pytest
from bloqade import (
    waveform,
    rydberg_h,
    piecewise_linear,
    piecewise_constant,
    constant,
    linear,
    var,
    cast,
    start,
    get_capabilities,
)
from bloqade.atom_arrangement import Chain
from bloqade.ir import (
    AnalogCircuit,
    Sequence,
    rydberg,
    Pulse,
    rabi,
    detuning,
    Field,
    Uniform,
)
from bloqade.ir.routine.base import Routine
from bloqade.ir.routine.params import Params
import numpy as np
from decimal import Decimal


def test_get_capabilities():
    capabilities = get_capabilities()

    assert capabilities.version == "0.6"
    assert capabilities.capabilities.task.number_shots_min == 1
    assert capabilities.capabilities.task.number_shots_max == 1000
    assert capabilities.capabilities.lattice.number_qubits_max == 256
    assert capabilities.capabilities.lattice.area.width == Decimal("75.0")
    assert capabilities.capabilities.lattice.area.height == Decimal("76.0")
    assert capabilities.capabilities.lattice.geometry.spacing_radial_min == Decimal(
        "4.0"
    )
    assert capabilities.capabilities.lattice.geometry.spacing_vertical_min == Decimal(
        "4.0"
    )
    assert capabilities.capabilities.lattice.geometry.position_resolution == Decimal(
        "0.1"
    )
    assert capabilities.capabilities.lattice.geometry.number_sites_max == 256
    assert capabilities.capabilities.rydberg.c6_coefficient == Decimal("5.42e6")
    assert capabilities.capabilities.rydberg.global_.rabi_frequency_min == Decimal(
        "0.0"
    )
    assert capabilities.capabilities.rydberg.global_.rabi_frequency_max == Decimal(
        "15.8"
    )
    assert (
        capabilities.capabilities.rydberg.global_.rabi_frequency_resolution
        == Decimal("4.0e-4")
    )
    assert (
        capabilities.capabilities.rydberg.global_.rabi_frequency_slew_rate_max
        == Decimal("250.0")
    )
    assert capabilities.capabilities.rydberg.global_.detuning_min == Decimal("-125.0")
    assert capabilities.capabilities.rydberg.global_.detuning_max == Decimal("125.0")
    assert capabilities.capabilities.rydberg.global_.detuning_resolution == Decimal(
        "2.0e-7"
    )
    assert capabilities.capabilities.rydberg.global_.detuning_slew_rate_max == Decimal(
        "2500.0"
    )
    assert capabilities.capabilities.rydberg.global_.phase_min == Decimal("-99.0")
    assert capabilities.capabilities.rydberg.global_.phase_max == Decimal("99.0")
    assert capabilities.capabilities.rydberg.global_.phase_resolution == Decimal(
        "5.0E-7"
    )
    assert capabilities.capabilities.rydberg.global_.time_min == Decimal("0.0")
    assert capabilities.capabilities.rydberg.global_.time_max == Decimal("4.0")
    assert capabilities.capabilities.rydberg.global_.time_resolution == Decimal("1e-3")
    assert capabilities.capabilities.rydberg.global_.time_delta_min == Decimal("5e-2")
    assert capabilities.capabilities.rydberg.local.detuning_min == Decimal("0.0")
    assert capabilities.capabilities.rydberg.local.detuning_max == Decimal("125.0")
    assert capabilities.capabilities.rydberg.local.detuning_slew_rate_max == Decimal(
        "1.2566e3"
    )
    assert capabilities.capabilities.rydberg.local.site_coefficient_min == Decimal(
        "0.0"
    )
    assert capabilities.capabilities.rydberg.local.site_coefficient_max == Decimal(
        "1.0"
    )
    assert (
        capabilities.capabilities.rydberg.local.number_local_detuning_sites
        == Decimal("200")
    )
    assert capabilities.capabilities.rydberg.local.spacing_radial_min == Decimal("5.0")
    assert capabilities.capabilities.rydberg.local.time_resolution == Decimal("1e-3")
    assert capabilities.capabilities.rydberg.local.time_delta_min == Decimal("5e-2")


def test_ir_piecewise_linear():
    A = piecewise_linear([0.1, 3.8, 0.2], [-10, -7, "a", "b"])

    ## Append type ir node
    assert len(A.waveforms) == 3
    assert A.waveforms[0].duration == cast(0.1)
    assert A.waveforms[0].start == cast(-10)
    assert A.waveforms[0].stop == cast(-7)

    assert A.waveforms[1].duration == cast(3.8)
    assert A.waveforms[1].start == cast(-7)
    assert A.waveforms[1].stop == cast("a")

    assert A.waveforms[2].duration == cast(0.2)
    assert A.waveforms[2].start == cast("a")
    assert A.waveforms[2].stop == cast("b")

    with pytest.raises(ValueError):
        piecewise_linear([0.1, 3.8, 0.2], [-10, -7, "a", "b", "c"])


def test_ir_const():
    A = constant(value=3.415, duration=0.55)

    ## Constant type ir node:
    assert A.value == cast(3.415)
    assert A.duration == cast(0.55)


def test_ir_linear():
    A = linear(start=0.5, stop=3.2, duration=0.76)

    ## Linear type ir node:
    assert A.start == cast(0.5)
    assert A.stop == cast(3.2)
    assert A.duration == cast(0.76)


def test_ir_piecewise_constant():
    A = piecewise_constant(durations=[0.1, 3.8, 0.2], values=[-10, "a", "b"])

    assert A.waveforms[0].duration == cast(0.1)
    assert A.waveforms[0].value == cast(-10)

    assert A.waveforms[1].duration == cast(3.8)
    assert A.waveforms[1].value == cast("a")

    assert A.waveforms[2].duration == cast(0.2)
    assert A.waveforms[2].value == cast("b")

    with pytest.raises(ValueError):
        piecewise_constant([0.1, 3.8, 0.2], [-10, -7, "a", "b"])


def test_rydberg_h():
    run_time = var("run_time")

    @waveform(run_time + 0.2)
    def delta(t, amp, omega):
        return np.sin(omega * t) * amp

    delta = delta.sample(0.05, "linear")
    ampl = piecewise_linear([0.1, run_time, 0.1], [0, 10, 10, 0])
    phase = piecewise_constant([2, 2], [0, np.pi])
    register = Chain(11, lattice_spacing=6.1)

    static_params = {"amp": 1.0}
    batch_params = [{"omega": omega} for omega in [1, 2, 4, 8]]
    args = ["run_time"]

    routine = rydberg_h(
        register,
        detuning=delta,
        amplitude=ampl,
        phase=phase,
        batch_params=batch_params,
        static_params=static_params,
        args=args,
    )

    detuning_field = Field({Uniform: delta})
    ampl_field = Field({Uniform: ampl})
    phase_field = Field({Uniform: phase})

    pulse = Pulse(
        {detuning: detuning_field, rabi.amplitude: ampl_field, rabi.phase: phase_field}
    )
    sequence = Sequence({rydberg: pulse})

    source = (
        register.rydberg.detuning.uniform.apply(delta)
        .amplitude.uniform.apply(ampl)
        .phase.uniform.apply(phase)
        .assign(**static_params)
        .batch_assign(batch_params)
        .args(args)
    )

    circuit = AnalogCircuit(register, sequence)
    params = Params(static_params, batch_params, args)
    expected_routine = Routine(source, circuit, params)

    # ignore because no equality implemented
    # assert routine.source == expected_routine.source
    assert routine.circuit == expected_routine.circuit
    assert routine.params == expected_routine.params


def test_rydberg_h_2():
    run_time = var("run_time")

    @waveform(run_time + 0.2)
    def delta(t, amp, omega):
        return np.sin(omega * t) * amp

    delta = delta.sample(0.05, "linear")
    ampl = piecewise_linear([0.1, run_time, 0.1], [0, 10, 10, 0])
    phase = piecewise_constant([2, 2], [0, np.pi])
    register = start.add_position((0, 0))

    prog = rydberg_h(
        (0, 0), detuning=delta, amplitude=ampl, phase=phase, batch_params={}
    )

    detuning_field = Field({Uniform: delta})
    ampl_field = Field({Uniform: ampl})
    phase_field = Field({Uniform: phase})

    pulse = Pulse(
        {detuning: detuning_field, rabi.amplitude: ampl_field, rabi.phase: phase_field}
    )
    sequence = Sequence({rydberg: pulse})

    print(prog.parse_circuit())
    print(AnalogCircuit(register, sequence))

    assert prog.parse_circuit() == AnalogCircuit(register, sequence)



from functools import reduce
from bloqade.builder.base import Builder
from bloqade.builder.typing import ScalarType
from bloqade.builder.route import WaveformRoute

from beartype import beartype
from beartype.typing import Optional, Union, List, Callable

import bloqade.ir as ir


class WaveformAttachable(Builder):
    @beartype
    def linear(
        self, start: ScalarType, stop: ScalarType, duration: ScalarType
    ) -> "Linear":
        """

        Append or assign a linear waveform to the current location(s).

        If you specified a spatial modulation (e.g. `uniform`, `location`,`scale`)
        previously without a waveform you will now have completed the construction
        of a "drive", one or a sum of drives creating a "field"
        (e.g. Real-valued Rabi Amplitude/Phase).

        If you have already specified a waveform previously you will now be appending
        this waveform to that previous waveform.

        ### Usage Example:
        ```
        >>> prog = start.add_position((0,0)).rydberg.detuning.uniform
        # apply a linear waveform that goes from 0 to 1 radians/us in 0.5 us
        >>> prog.linear(start=0,stop=1,duration=0.5)
        ```

        - Your next steps include:
        - Continue building your waveform via:
            - `...linear(start, stop, duration).linear(start, stop, duration)`:
                to append another linear waveform
            - `...linear(start, stop, duration).constant(value, duration)`:
                to append a constant waveform
            - `...linear(start, stop, duration)
                .piecewise_linear([durations], [values])`:
                to append a piecewise linear waveform
            - `...linear(start, stop, duration)
                .piecewise_constant([durations], [values])`:
                to append a piecewise constant waveform
            - `...linear(start, stop, duration).poly([coefficients], duration)`:
                to append a polynomial waveform
            - `...linear(start, stop, duration).apply(wf:bloqade.ir.Waveform)`:
                to append a pre-defined waveform
            - `...linear(start, stop, duration).fn(f(t,...))`:
                to append a waveform defined by a python function
        - Slice a portion of the waveform to be used:
            - `...linear(start, stop, duration).slice(start, stop, duration)`
        - Save the ending value of your waveform to be reused elsewhere
            - `...linear(start, stop, duration).record("you_variable_here")`
        - Begin constructing another drive by starting a new spatial modulation
            (this drive will be summed to the one you just created):
            - `...linear(start, stop, duration).uniform`:
                To address all atoms in the field
            - `...linear(start, stop, duration).location(int)`:
                To address atoms at specific location with scaling
            - `...linear(start, stop, duration).scale(...)`
                - To address atoms at specific location with scaling
                - To address multiple atoms at specific locations by specifying
                    a single variable and then assigning it a list of coordinates
        - Assign values to pre-existing variables via:
            - `...linear(start, stop, duration).assign(variable_name = value)`:
                to assign a single value to a variable
            - `...linear(start, stop, duration)
                .batch_assign(variable_name = [value1, ...])`:
                to assign multiple values to a variable
            - `...linear(start, stop, duration).args(["previously_defined_var"])`:
                to defer assignment of a variable to execution time
        - Select the backend you want your program to run on via:
            - `...linear(start, stop, duration).braket`:
                to run on Braket local emulator or QuEra hardware remotely
            - `...linear(start, stop, duration).bloqade`:
                to run on the Bloqade local emulator
            - `...linear(start, stop, duration).device`:
                to specify the backend via string
        - Choose to parallelize your atom geometry,
          duplicating it to fill the whole space:
            - `...linear(start, stop, duration).parallelize(spacing)`
        - Start targeting another level coupling
            - `...linear(start, stop, duration).rydberg`:
                to target the Rydberg level coupling
            - `...linear(start, stop, duration).hyperfine`:
                to target the Hyperfine level coupling
        - Start targeting other fields within your current level coupling
          (previously selected as `rydberg` or `hyperfine`):
            - `...linear(start, stop, duration).amplitude`:
                to target the real-valued Rabi Amplitude field
            - `...linear(start, stop, duration).phase`:
                to target the real-valued Rabi Phase field
            - `...linear(start, stop, duration).detuning`:
                to target the Detuning field
            - `...linear(start, stop, duration).rabi`:
                to target the complex-valued Rabi field
        """

        return Linear(start, stop, duration, self)

    @beartype
    def constant(self, value: ScalarType, duration: ScalarType) -> "Constant":
        """
        Append or assign a constant waveform to the current location(s).

        If you specified a spatial modulation (e.g. `uniform`, `location`,`scale`)
        previously without a waveform you will now have completed the construction
        of a "drive", one or a sum of drives creating a "field"
        (e.g. Real-valued Rabi Amplitude/Phase).

        If you have already specified a waveform previously you will now be appending
        this waveform to that previous waveform.

        ### Usage Example:
        ```
        >>> prog = start.add_position((0,0)).rydberg.detuning.uniform
        # apply a constant waveform of 1.9 radians/us for 0.5 us
        >>> prog.constant(value=1.9,duration=0.5)
        ```

        - Your next steps include:
        - Continue building your waveform via:
            - `...constant(value, duration).linear(start, stop, duration)`:
                to append another linear waveform
            - `...constant(value, duration).constant(value, duration)`:
                to append a constant waveform
            - `...constant(value, duration)
                .piecewise_linear([durations], [values])`:
                to append a piecewise linear waveform
            - `...constant(value, duration)
                .piecewise_constant([durations], [values])`:
                to append a piecewise constant waveform
            - `...constant(value, duration).poly([coefficients], duration)`:
                to append a polynomial waveform
            - `...constant(value, duration).apply(wf:bloqade.ir.Waveform)`:
                to append a pre-defined waveform
            - `...constant(value, duration).fn(f(t,...))`:
                to append a waveform defined by a python function
        - Slice a portion of the waveform to be used:
            - `...constant(value, duration).slice(start, stop, duration)`
        - Save the ending value of your waveform to be reused elsewhere
            - `...constant(value, duration).record("you_variable_here")`
        - Begin constructing another drive by starting a new spatial modulation
            (this drive will be summed to the one you just created):
            - `...constant(value, duration).uniform`:
                To address all atoms in the field
            - `...constant(value, duration).scale(...)`:
                To address an atom at a specific location via index
            - `...constant(value, duration).location(int)`
                - To address an atom at a specific location via variable
                - To address multiple atoms at specific locations by specifying
                    a single variable and then assigning it a list of coordinates
        - Assign values to pre-existing variables via:
            - `...constant(value, duration).assign(variable_name = value)`:
                to assign a single value to a variable
            - `...constant(value, duration)
                .batch_assign(variable_name = [value1, ...])`:
                to assign multiple values to a variable
            - `...constant(value, duration).args(["previously_defined_var"])`:
                to defer assignment of a variable to execution time
        - Select the backend you want your program to run on via:
            - `...constant(value, duration).braket`:
                to run on Braket local emulator or QuEra hardware remotely
            - `...constant(value, duration).bloqade`:
                to run on the Bloqade local emulator
            - `...constant(value, duration).device`:
                to specify the backend via string
        - Choose to parallelize your atom geometry,
          duplicating it to fill the whole space:
            - `...constant(start, stop, duration).parallelize(spacing)`
        - Start targeting another level coupling
            - `...constant(value, duration).rydberg`:
                to target the Rydberg level coupling
            - `...constant(value, duration).hyperfine`:
                to target the Hyperfine level coupling
        - Start targeting other fields within your current
          level coupling (previously selected as `rydberg` or `hyperfine`):
            - `...constant(value, duration).amplitude`:
                to target the real-valued Rabi Amplitude field
            - `...constant(value, duration).phase`:
                to target the real-valued Rabi Phase field
            - `...constant(value, duration).detuning`:
                to target the Detuning field
            - `...constant(value, duration).rabi`:
                to target the complex-valued Rabi field

        """
        return Constant(value, duration, self)

    @beartype
    def poly(self, coeffs: List[ScalarType], duration: ScalarType) -> "Poly":
        """
        Append or assign a waveform with a polynomial profile to current location(s).

        You pass in a list of coefficients and a duration to this method which obeys
        the following expression:

        `
        wv(t) = coeffs[0] + coeffs[1]*t + coeffs[2]*t^2 + ... + coeffs[n]*t^n
        `

        If you specified a spatial modulation (e.g. `uniform`, `location`,`scale`)
        previously without a waveform you will now have completed the construction
        of a "drive", one or a sum of drives creating a "field"
        (e.g. Real-valued Rabi Amplitude/Phase).

        If you have already specified a waveform previously you will now be appending
        this waveform to that previous waveform.

        ### Usage Example:
        ```
        >>> prog = start.add_position((0,0)).rydberg.detuning.uniform
        >>> coeffs = [-1, 0.5, 1.2]
        # resulting polynomial is:
        # f(t) = -1 + 0.5*t + 1.2*t^2 with duration of
        # 0.5 us
        >>> prog.poly(coeffs, duration=0.5)
        ```

        - Your next steps include:
        - Continue building your waveform via:
            - `...poly([coeffs], duration).linear(start, stop, duration)`:
                to append another linear waveform
            - `...poly([coeffs], duration).constant(value, duration)`:
                to append a constant waveform
            - `...poly([coeffs], duration)
                .piecewise_linear([durations], [values])`:
                to append a piecewise linear waveform
            - `...poly([coeffs], duration)
                .piecewise_constant([durations],[values])`:
                to append a piecewise constant waveform
            - `...poly([coeffs], duration).poly([coefficients], duration)`:
                to append a polynomial waveform
            - `...poly([coeffs], duration).apply(waveform)`:
                to append a pre-defined waveform
            - `...poly([coeffs], duration).fn(f(t,...))`:
                to append a waveform defined by a python function
        - Slice a portion of the waveform to be used:
            - `...poly([coeffs], duration).slice(start, stop, duration)`
        - Save the ending value of your waveform to be reused elsewhere
            - `...poly([coeffs], duration).record("you_variable_here")`
        - Begin constructing another drive by starting a new spatial modulation
          (this drive will be summed to the one you just created):
            - `...poly([coeffs], duration).uniform`:
                To address all atoms in the field
            - `...poly([coeffs], duration).location(int)`:
                To address an atom at a specific location via index
            - `...poly([coeffs], duration).scale(...)`
                - To address an atom at a specific location via variable
                - To address multiple atoms at specific locations by
                    specifying a single variable and then assigning
                    it a list of coordinates
        - Assign values to pre-existing variables via:
            - `...poly([coeffs], duration).assign(variable_name = value)`:
                to assign a single value to a variable
            - `...poly([coeffs], duration)
                .batch_assign(variable_name = [value1, ...])`:
                to assign multiple values to a variable
            - `...poly([coeffs], duration).args(["previously_defined_var"])`:
                to defer assignment of a variable to execution time
        - Select the backend you want your program to run on via:
            - `...poly([coeffs], duration).braket`:
                to run on Braket local emulator or QuEra hardware remotely
            - `...poly([coeffs], duration).bloqade`:
                to run on the Bloqade local emulator
            - `...poly([coeffs], duration).device`:
                to specify the backend via string
        - Choose to parallelize your atom geometry,
          duplicating it to fill the whole space:
            - `...poly([coeffs], duration).parallelize(spacing)`
        - Start targeting another level coupling
            - `...poly([coeffs], duration).rydberg`:
                to target the Rydberg level coupling
            - `...poly([coeffs], duration).hyperfine`:
                to target the Hyperfine level coupling
        - Start targeting other fields within your current level
          coupling (previously selected as `rydberg` or `hyperfine`):
            - `...poly([coeffs], duration).amplitude`:
                to target the real-valued Rabi Amplitude field
            - `...poly([coeffs], duration).phase`:
                to target the real-valued Rabi Phase field
            - `...poly([coeffs], duration).detuning`:
                to target the Detuning field
            - `...poly([coeffs], duration).rabi`:
                to target the complex-valued Rabi field
        """
        return Poly(coeffs, duration, self)

    @beartype
    def apply(self, wf: ir.Waveform) -> "Apply":
        """
        Apply a [`Waveform`][bloqade.ir.control.Waveform] built previously to
        current location(s).

        If you specified a spatial modulation (e.g. `uniform`, `location`,`scale`)
        previously without a waveform you will now have completed the construction
        of a "drive", one or a sum of drives creating a "field"
        (e.g. Real-valued Rabi Amplitude/Phase).

        If you have already specified a waveform previously you will now be appending
        this waveform to that previous waveform.

        ### Usage Example:
        ```
        >>> prog = start.add_position((0,0)).rydberg.detuning.uniform
        # build our waveform independently of the main program
        >>> from bloqade import piecewise_linear
        >>> wf = piecewise_linear(durations=[0.3, 2.5, 0.3],
        values=[0.0, 2.0, 2.0, 0.0])
        >>> prog.apply(wf)
        ```

        - Your next steps include:
        - Continue building your waveform via:
            - `...apply(waveform).linear(start, stop, duration)`:
                to append another linear waveform
            - `...apply(waveform).constant(value, duration)`:
                to append a constant waveform
            - `...apply(waveform).piecewise_linear([durations], [values])`:
                to append a piecewise linear waveform
            - `...apply(waveform).piecewise_constant([durations], [values])`:
                to append a piecewise constant waveform
            - `...apply(waveform).poly([coefficients], duration)`:
                to append a polynomial waveform
            - `...apply(waveform).apply(waveform)`:
                to append a pre-defined waveform
            - `...apply(waveform).fn(f(t,...))`:
                to append a waveform defined by a python function
        - Slice a portion of the waveform to be used:
            - `...apply(waveform).slice(start, stop, duration)`
        - Save the ending value of your waveform to be reused elsewhere
            - `...apply(waveform).record("you_variable_here")`
        - Begin constructing another drive by starting a new spatial modulation
          (this drive will be summed to the one you just created):
            - `...apply(waveform).uniform`: To address all atoms in the field
            - `...apply(waveform).location(int)`:
                To address an atom at a specific location via index
            - `...apply(waveform).scale(...)`
                - To address an atom at a specific location via variable
                - To address multiple atoms at specific locations by specifying a
                    single variable and then assigning it a list of coordinates
        - Assign values to pre-existing variables via:
            - `...apply(waveform).assign(variable_name = value)`:
                to assign a single value to a variable
            - `...apply(waveform).batch_assign(variable_name = [value1, ...])`:
                to assign multiple values to a variable
            - `...apply(waveform).args(["previously_defined_var"])`:
                to defer assignment of a variable to execution time
        - Select the backend you want your program to run on via:
            - `...apply(waveform).braket`:
                to run on Braket local emulator or QuEra hardware remotely
            - `...apply(waveform).bloqade`:
                to run on the Bloqade local emulator
            - `...apply(waveform).device`:
                to specify the backend via string
        - Choose to parallelize your atom geometry,
          duplicating it to fill the whole space:
            - `...apply(waveform).parallelize(spacing)`
        - Start targeting another level coupling
            - `...apply(waveform).rydberg`: to target the Rydberg level coupling
            - `...apply(waveform).hyperfine`: to target the Hyperfine level coupling
        - Start targeting other fields within your current level coupling
          (previously selected as `rydberg` or `hyperfine`):
            - `...apply(waveform).amplitude`:
                to target the real-valued Rabi Amplitude field
            - `...apply(waveform).phase`:
                to target the real-valued Rabi Phase field
            - `...apply(waveform).detuning`:
                to target the Detuning field
            - `...apply(waveform).rabi`:
                to target the complex-valued Rabi field
        """
        return Apply(wf, self)

    @beartype
    def piecewise_linear(
        self, durations: List[ScalarType], values: List[ScalarType]
    ) -> "PiecewiseLinear":
        """
        Append or assign a piecewise linear waveform to current location(s),
        where the waveform is formed by connecting `values[i], values[i+1]`
        with linear segments.

        The `durations` argument should have # of elements = len(values) - 1.
        `durations` should be the duration PER section of the waveform, NON-CUMULATIVE.

        If you specified a spatial modulation (e.g. `uniform`, `location`,`scale`)
        previously without a waveform you will now have completed the construction
        of a "drive", one or a sum of drives creating a "field"
        (e.g. Real-valued Rabi Amplitude/Phase).

        If you have already specified a waveform previously you will now be appending
        this waveform to that previous waveform.

        ### Usage Example:
        ```
        >>> prog = start.add_position((0,0)).rydberg.detuning.uniform
        # ramp our waveform up to a certain value, hold it
        # then ramp down. In this case, we ramp up to 2.0 rad/us in 0.3 us,
        # then hold it for 1.5 us before ramping down in 0.3 us back to 0.0 rad/us.
        >>> prog.piecewise_linear(durations=[0.3, 2.0, 0.3],
        values=[0.0, 2.0, 2.0, 0.0])
        ```

        - Your next steps include:
        - Continue building your waveform via:
            - `...piecewise_linear([durations], [values])
                .linear(start, stop, duration)`:
                to append another linear waveform
            - `...piecewise_linear([durations], [values]).constant(value, duration)`:
                to append a constant waveform
            - `...piecewise_linear([durations], [values])
                .piecewise_linear(durations, values)`:
                to append a piecewise linear waveform
            - `...piecewise_linear([durations], [values])
                .piecewise_constant([durations], [values])`:
                to append a piecewise constant waveform
            - `...piecewise_linear([durations], [values])
                .poly([coefficients], duration)`: to append a polynomial waveform
            - `...piecewise_linear([durations], [values]).apply(waveform)`:
                to append a pre-defined waveform
            - `...piecewise_linear([durations], [values]).fn(f(t,...))`:
                to append a waveform defined by a python function
        - Slice a portion of the waveform to be used:
            - `...piecewise_linear([durations], [values])
                .slice(start, stop, duration)`
        - Save the ending value of your waveform to be reused elsewhere
            - `...piecewise_linear([durations], [values])
                .record("you_variable_here")`
        - Begin constructing another drive by starting a new spatial modulation
          (this drive will be summed to the one you just created):
            - `...piecewise_linear([durations], [values]).uniform`:
                To address all atoms in the field
            - `...piecewise_linear([durations], [values]).scale(...)`:
                To address an atom at a specific location via index
            - `...piecewise_linear([durations], [values]).location(int)`
                - To address an atom at a specific location via variable
                - To address multiple atoms at specific locations by
                    specifying a single variable and then assigning it a
                    list of coordinates
        - Assign values to pre-existing variables via:
            - `...piecewise_linear([durations], [values])
                .assign(variable_name = value)`:
                to assign a single value to a variable
            - `...piecewise_linear([durations], [values])
                .batch_assign(variable_name = [value1, ...])`:
                to assign multiple values to a variable
            - `...piecewise_linear([durations], [values])
                .args(["previously_defined_var"])`:
                to defer assignment of a variable to execution time
        - Select the backend you want your program to run on via:
            - `...piecewise_linear([durations], [values]).braket`:
                to run on Braket local emulator or QuEra hardware remotely
            - `...piecewise_linear([durations], [values]).bloqade`:
                to run on the Bloqade local emulator
            - `...piecewise_linear([durations], [values]).device`:
                to specify the backend via string
        - Choose to parallelize your atom geometry,
          duplicating it to fill the whole space:
            - `...piecewise_linear([durations], [values]).parallelize(spacing)`
        - Start targeting another level coupling
            - `...piecewise_linear([durations], [values]).rydberg`:
                to target the Rydberg level coupling
            - `...piecewise_linear([durations], [values]).hyperfine`:
                to target the Hyperfine level coupling
        - Start targeting other fields within your current level coupling
          (previously selected as `rydberg` or `hyperfine`):
            - `...piecewise_linear([durations], [values]).amplitude`:
                to target the real-valued Rabi Amplitude field
            - `...piecewise_linear([durations], [values]).phase`:
                to target the real-valued Rabi Phase field
            - `...piecewise_linear([durations], [values]).detuning`:
                to target the Detuning field
            - `....rabi`: to target the complex-valued Rabi field
        """
        return PiecewiseLinear(durations, values, self)

    @beartype
    def piecewise_constant(
        self, durations: List[ScalarType], values: List[ScalarType]
    ) -> "PiecewiseConstant":
        """
        Append or assign a piecewise constant waveform to current location(s).

        The `durations` argument should have number of elements = len(values).
        `durations` should be the duration PER section of the waveform,
        NON-CUMULATIVE.

        If you specified a spatial modulation (e.g. `uniform`, `location`,`scale`)
        previously without a waveform you will now have completed the construction
        of a "drive", one or a sum of drives creating a "field"
        (e.g. Real-valued Rabi Amplitude/Phase).

        If you have already specified a waveform previously you will now be appending
        this waveform to that previous waveform.

        ### Usage Example:
        ```
        >>> prog = start.add_position((0,0)).rydberg.rabi.phase.uniform
        # create a staircase, we hold 0.0 rad/us for 1.0 us, then
        # to 1.0 rad/us for 0.5 us before stopping at 0.8 rad/us for 0.9 us.
        >>> prog.piecewise_linear(durations=[0.3, 2.0, 0.3], values=[1.0, 0.5, 0.9])
        ```

        - Your next steps including:
        - Continue building your waveform via:
            - `...piecewise_constant([durations], [values])
                .linear(start, stop, duration)`: to append another linear waveform
            - `...piecewise_constant([durations], [values])
                .constant(value, duration)`: to append a constant waveform
            - `...piecewise_constant([durations], [values])
                .piecewise_linear([durations], [values])`:
                to append a piecewise linear waveform
            - `...piecewise_constant([durations], [values])
                .piecewise_constant([durations], [values])`:
                to append a piecewise constant waveform
            - `...piecewise_constant([durations], [values])
                .poly([coefficients], duration)`: to append a polynomial waveform
            - `...piecewise_constant([durations], [values])
                .apply(waveform)`: to append a pre-defined waveform
            - `...piecewise_constant([durations], [values]).fn(f(t,...))`:
                to append a waveform defined by a python function
        - Slice a portion of the waveform to be used:
            - `...piecewise_constant([durations], [values])
                .slice(start, stop, duration)`
        - Save the ending value of your waveform to be reused elsewhere
            - `...piecewise_constant([durations], [values])
                .record("you_variable_here")`
        - Begin constructing another drive by starting a new spatial modulation
          (this drive will be summed to the one you just created):
            - `...piecewise_constant([durations], [values]).uniform`:
                To address all atoms in the field
            - `...piecewise_constant([durations], [values]).location(int)`:
                To address an atom at a specific location via index
            - `...piecewise_constant([durations], [values]).scale(...)`
                - To address an atom at a specific location via variable
                - To address multiple atoms at specific locations by
                    specifying a single variable and then assigning it a
                    list of coordinates
        - Assign values to pre-existing variables via:
            - `...piecewise_constant([durations], [values])
                .assign(variable_name = value)`: to assign a single value to a variable
            - `...piecewise_constant([durations], [values])
                .batch_assign(variable_name = [value1, ...])`:
                to assign multiple values to a variable
            - `...piecewise_constant([durations], [values])
                .args(["previously_defined_var"])`: to defer assignment
                of a variable to execution time
        - Select the backend you want your program to run on via:
            - `...piecewise_constant([durations], [values]).braket`:
                to run on Braket local emulator or QuEra hardware remotely
            - `...piecewise_constant([durations], [values]).bloqade`:
                to run on the Bloqade local emulator
            - `...piecewise_constant([durations], [values]).device`:
                to specify the backend via string
        - Choose to parallelize your atom geometry,
          duplicating it to fill the whole space:
            - `...piecewise_constat([durations], [values]).parallelize(spacing)`
        - Start targeting another level coupling
            - `...piecewise_constant([durations], [values]).rydberg`:
                to target the Rydberg level coupling
            - `...piecewise_constant([durations], [values]).hyperfine`:
                to target the Hyperfine level coupling
        - Start targeting other fields within your current level coupling
          (previously selected as `rydberg` or `hyperfine`):
            - `...piecewise_constant(durations, values).amplitude`:
                to target the real-valued Rabi Amplitude field
            - `...piecewise_constant([durations], [values]).phase`:
                to target the real-valued Rabi Phase field
            - `...piecewise_constant([durations], [values]).detuning`:
                to target the Detuning field
            - `...piecewise_constant([durations], [values]).rabi`:
                to target the complex-valued Rabi field
        """
        return PiecewiseConstant(durations, values, self)

    @beartype
    def fn(self, fn: Callable, duration: ScalarType) -> "Fn":
        """
        Append or assign a custom function as a waveform.

        The function must have its first argument be that of time but
        can also have other arguments which are treated as variables.
        You can assign values to later in the program via `.assign` or `.batch_assign`.

        The function must also return a singular float value.

        If you specified a spatial modulation (e.g. `uniform`, `location`,`scale`)
        previously without a waveform you will now have completed the construction
        of a "drive", one or a sum of drives creating a "field"
        (e.g. Real-valued Rabi Amplitude/Phase).

        If you have already specified a waveform previously you will now be appending
        this waveform to that previous waveform.

        ### ### Usage Examples:
        ```
        >>> prog = start.add_position((0,0)).rydberg.detuning.uniform
        # define our custom waveform. It must have one argument
        # be time followed by any other number of arguments that can
        # be assigned a value later in the program via `.assign` or `.batch_assign`
        >>> def custom_waveform_function(t, arg1, arg2):
                return arg1*t + arg2
        >>> prog = prog.fn(custom_waveform_function, duration = 0.5)
        # assign values
        >>> assigned_vars_prog = prog.assign(arg1 = 1.0, arg2 = 2.0)
        # or go for batching!
        >>> assigned_vars_batch_prog = prog.assign(arg1 = 1.0, arg2 = [1.0, 2.0, 3.0])
        ```

        - Your next steps include:
        - Continue building your waveform via:
            - `...fn(f(t,...))
                .linear(start, stop, duration)`: to append another linear waveform
            - `...fn(f(t,...))
                .constant(value, duration)`: to append a constant waveform
            - `...fn(f(t,...))
                .piecewise_linear(durations, values)`:
                to append a piecewise linear waveform
            - `...fn(f(t,...))
                .piecewise_constant(durations, values)`:
                to append a piecewise constant waveform
            - `...fn(f(t,...))
                .poly([coefficients], duration)`: to append a polynomial waveform
            - `...fn(f(t,...))
                .apply(waveform)`: to append a pre-defined waveform
            - `...fn(f(t,...))
                .fn(f(t,...))`: to append a waveform defined by a python function
        - Slice a portion of the waveform to be used:
            - `...fn(f(t,...)).slice(start, stop, duration)`
        - Save the ending value of your waveform to be reused elsewhere
            - `...fn(f(t,...)).record("you_variable_here")`
        - Begin constructing another drive by starting a new spatial modulation
          (this drive will be summed to the one you just created):
            - `...fn(f(t,...)).uniform`:
                To address all atoms in the field
            - `...fn(f(t,...)).scale(...)`:
                To address an atom at a specific location via index
            - ...fn(f(t,...)).location(int)`
                - To address an atom at a specific location via variable
                - To address multiple atoms at specific locations by
                    specifying a single variable and then assigning it a
                    list of coordinates
        - Assign values to pre-existing variables via:
            - `...fn(f(t,...))
                .assign(variable_name = value)`: to assign a single value to a variable
            - `...fn(f(t,...))
                .batch_assign(variable_name = [value1, ...])`:
                to assign multiple values to a variable
            - `...fn(f(t,...))
                .args(["previously_defined_var"])`:
                to defer assignment of a variable to execution time
        - Select the backend you want your program to run on via:
            - `...fn(f(t,...)).braket`:
                to run on Braket local emulator or QuEra hardware remotely
            - `...fn(f(t,...)).bloqade`:
                to run on the Bloqade local emulator
            - `...fn(f(t,...)).device`:
                to specify the backend via string
        - Choose to parallelize your atom geometry,
          duplicating it to fill the whole space:
            - `...fn(f(t,...)).parallelize(spacing)`
        - Start targeting another level coupling
            - `...fn(f(t,...)).rydberg`:
                to target the Rydberg level coupling
            - `...fn(f(t,...)).hyperfine`:
                to target the Hyperfine level coupling
        - Start targeting other fields within your current level coupling
          (previously selected as `rydberg` or `hyperfine`):
            - `...fn(f(t,...)).amplitude`:
                to target the real-valued Rabi Amplitude field
            - `...fn(f(t,...)).phase`:
                to target the real-valued Rabi Phase field
            - `...fn(f(t,...)).detuning`:
                to target the Detuning field
            - `...fn(f(t,...)).rabi`:
                to target the complex-valued Rabi field

        """
        return Fn(fn, duration, self)


# NOTE: waveform can refer previous pulse notes
#       or continue to specify pragma options
class Waveform(WaveformRoute, WaveformAttachable):
    pass


# mixin for slice and record
class Sliceable:
    @beartype
    def slice(
        self,
        start: Optional[ScalarType] = None,
        stop: Optional[ScalarType] = None,
    ) -> "Slice":
        """
        Indicate that you only want a portion of your waveform to be used in
        the program.

        If you specified a spatial modulation (e.g. `uniform`, `location`,`scale`)
        previously without a waveform you will now have completed the construction
        of a "drive", one or a sum of drives creating a "field"
        (e.g. Real-valued Rabi Amplitude/Phase).

        If you have already specified a waveform previously you will now be appending
        this waveform to that previous waveform.


        ### Usage Example:
        ```
        # define a program with a waveform of interest
        >>> from bloqade import start
        >>> prog = start.add_position((0,0)).rydberg.rabi.amplitude.uniform
        >>> prog_with_wf = prog.piecewise_linear(durations=[0.3, 2.0, 0.3],
        values=[0.0, 2.0, 2.0, 0.0])
        # instead of using the full waveform we opt to only take the first 1 us
        >>> prog_with_slice = prog_with_wf.slice(0.0, 1.0)
        # you may use variables as well
        >>> prog_with_slice = prog_with_wf.slice("start", "end")
        ```

        - Your next steps include:
        - Continue building your waveform via:
            - `...slice(start, stop).linear(start, stop, duration)`:
                to append another linear waveform
            - `...slice(start, stop).constant(value, duration)`:
                to append a constant waveform
            - `...slice(start, stop).piecewise_linear()`:
                to append a piecewise linear waveform
            - `...slice(start, stop).piecewise_constant()`:
                to append a piecewise constant waveform
            - `...slice(start, stop).poly([coefficients], duration)`:
                to append a polynomial waveform
            - `...slice(start, stop).apply(wf:bloqade.ir.Waveform)`:
                to append a pre-defined waveform
            - `...slilce(start, stop).fn(f(t,...))`:
                to append a waveform defined by a python function
        - Begin constructing another drive by starting a new spatial modulation
            (this drive will be summed to the one you just created):
            - `...slice(start, stop).uniform`:
                To address all atoms in the field
            - `...slice(start, stop).location(int)`:
                To address an atom at a specific location via index
            - `...slice(start, stop).scale(...)`
                - To address an atom at a specific location via variable
                - To address multiple atoms at specific locations by specifying
                    a single variable and then assigning it a list of coordinates
        - Assign values to pre-existing variables via:
            - `...slice(start, stop).assign(variable_name = value)`:
                to assign a single value to a variable
            - `...slice(start, stop)
                .batch_assign(variable_name = [value1, ...])`:
                to assign multiple values to a variable
            - `...slice(start, stop).args(["previously_defined_var"])`:
                to defer assignment of a variable to execution time
        - Select the backend you want your program to run on via:
            - `...slice(start, stop).braket`:
                to run on Braket local emulator or QuEra hardware remotely
            - `...slice(start, stop).bloqade`:
                to run on the Bloqade local emulator
            - `...slice(start, stop).device`:
                to specify the backend via string
        - Choose to parallelize your atom geometry,
          duplicating it to fill the whole space:
            - `...slice(start, stop).parallelize(spacing)`
        - Start targeting another level coupling
            - `...slice(start, stop).rydberg`:
                to target the Rydberg level coupling
            - `...slice(start, stop).hyperfine`:
                to target the Hyperfine level coupling
        - Start targeting other fields within your current level coupling
          (previously selected as `rydberg` or `hyperfine`):
            - `...slice(start, stop).amplitude`:
                to target the real-valued Rabi Amplitude field
            - `...slice(start, stop).phase`:
                to target the real-valued Rabi Phase field
            - `...slice(start, stop).detuning`:
                to target the Detuning field
            - `...slice(start, stop).rabi`:
                to target the complex-valued Rabi field
        """
        return Slice(start, stop, self)


class Recordable:
    @beartype
    def record(self, name: str) -> "Record":
        """
        Copy or "record" the value at the end of the waveform into a variable
        so that it can be used in another place.

        A common design pattern is to couple this with `.slice()` considering
        you may not know exactly what the end value of a `.slice()` is,
        especially in parameter sweeps where it becomes cumbersome to handle.

        If you specified a spatial modulation (e.g. `uniform`, `location`,`scale`)
        previously without a waveform you will now have completed the construction
        of a "drive", one or a sum of drives creating a "field"
        (e.g. Real-valued Rabi Amplitude/Phase).

        If you have already specified a waveform previously you will now be appending
        this waveform to that previous waveform.

        ### Usage Example:
        ```
        # define program of interest
        >>> from bloqade import start
        >>> prog = start.rydberg.rabi.amplitude.uniform
        >>> prog_with_wf = prog.piecewise_linear(durations=[0.3, 2.0, 0.3],
        values=[0.0, 2.0, 2.0, 0.0])
        # We now slice the piecewise_linear from above and record the
        # value at the end of that slice. We then use that value
        # to construct a new waveform that can be appended to the previous
        # one without introducing discontinuity (refer to the
        # "Quantum Scar Dynamics" tutorial for how this could be handy)
        >>> prog_with_record = prog_with_wf.slice(0.0, 1.0).record("end_of_wf")
        >>> record_applied_prog = prog_with_record.linear(start="end_of_wf"
        , stop=0.0, duration=0.3)
        ```

        - Your next steps include:
        - Continue building your waveform via:
            - `...slice(start, stop).linear(start, stop, duration)`:
                to append another linear waveform
            - `...slice(start, stop).constant(value, duration)`:
                to append a constant waveform
            - `...slice(start, stop).piecewise_linear()`:
                to append a piecewise linear waveform
            - `...slice(start, stop).piecewise_constant()`:
                to append a piecewise constant waveform
            - `...slice(start, stop).poly([coefficients], duration)`:
                to append a polynomial waveform
            - `...slice(start, stop).apply(wf:bloqade.ir.Waveform)`:
                to append a pre-defined waveform
            - `...slilce(start, stop).fn(f(t,...))`:
                to append a waveform defined by a python function
        - Begin constructing another drive by starting a new spatial modulation
            (this drive will be summed to the one you just created):
            - `...slice(start, stop).uniform`:
                To address all atoms in the field
            - `...slice(start, stop).location(int)`:
                To address an atom at a specific location via index
            - `...slice(start, stop).scale(str)`
                - To address an atom at a specific location via variable
                - To address multiple atoms at specific locations by specifying
                    a single variable and then assigning it a list of coordinates
        - Assign values to pre-existing variables via:
            - `...slice(start, stop).assign(variable_name = value)`:
                to assign a single value to a variable
            - `...slice(start, stop)
                .batch_assign(variable_name = [value1, ...])`:
                to assign multiple values to a variable
            - `...slice(start, stop).args(["previously_defined_var"])`:
                to defer assignment of a variable to execution time
        - Select the backend you want your program to run on via:
            - `...slice(start, stop).braket`:
                to run on Braket local emulator or QuEra hardware remotely
            - `...slice(start, stop).bloqade`:
                to run on the Bloqade local emulator
            - `...slice(start, stop).device`:
                to specify the backend via string
        - Choose to parallelize your atom geometry,
          duplicating it to fill the whole space:
            - `...slice(start, stop).parallelize(spacing)`
        - Start targeting another level coupling
            - `...slice(start, stop).rydberg`:
                to target the Rydberg level coupling
            - `...slice(start, stop).hyperfine`:
                to target the Hyperfine level coupling
        - Start targeting other fields within your current level coupling
          (previously selected as `rydberg` or `hyperfine`):
            - `...slice(start, stop).amplitude`:
                to target the real-valued Rabi Amplitude field
            - `...slice(start, stop).phase`:
                to target the real-valued Rabi Phase field
            - `...slice(start, stop).detuning`:
                to target the Detuning field
            - `...slice(start, stop).rabi`:
                to target the complex-valued Rabi field
        ```
        """
        return Record(name, self)


class WaveformPrimitive(Waveform, Sliceable, Recordable):
    def __bloqade_ir__(self):
        raise NotImplementedError


class Linear(WaveformPrimitive):
    def __init__(
        self,
        start: ScalarType,
        stop: ScalarType,
        duration: ScalarType,
        parent: Optional[Builder] = None,
    ) -> None:
        super().__init__(parent)
        self._start = ir.cast(start)
        self._stop = ir.cast(stop)
        self._duration = ir.cast(duration)

    def __bloqade_ir__(self) -> ir.Linear:
        return ir.Linear(start=self._start, stop=self._stop, duration=self._duration)


class Constant(WaveformPrimitive):
    def __init__(
        self, value: ScalarType, duration: ScalarType, parent: Optional[Builder] = None
    ) -> None:
        super().__init__(parent)
        self._value = ir.cast(value)
        self._duration = ir.cast(duration)

    def __bloqade_ir__(self) -> ir.Constant:
        return ir.Constant(value=self._value, duration=self._duration)


class Poly(WaveformPrimitive):
    def __init__(
        self,
        coeffs: List[ScalarType],
        duration: ScalarType,
        parent: Optional[Builder] = None,
    ) -> None:
        super().__init__(parent)
        self._coeffs = ir.cast(coeffs)
        self._duration = ir.cast(duration)

    def __bloqade_ir__(self):
        return ir.Poly(coeffs=self._coeffs, duration=self._duration)


class Apply(WaveformPrimitive):
    def __init__(self, wf: ir.Waveform, parent: Optional[Builder] = None):
        super().__init__(parent)
        self._wf = wf

    def __bloqade_ir__(self):
        return self._wf


class PiecewiseLinear(WaveformPrimitive):
    def __init__(
        self,
        durations: List[ScalarType],
        values: List[ScalarType],
        parent: Optional[Builder] = None,
    ):
        assert (
            len(durations) == len(values) - 1
        ), "durations and values must be the same length"

        super().__init__(parent)
        self._durations = list(map(ir.cast, durations))
        self._values = list(map(ir.cast, values))

    def __bloqade_ir__(self):
        iter = zip(self._values[:-1], self._values[1:], self._durations)
        wfs = [ir.Linear(start=v0, stop=v1, duration=t) for v0, v1, t in iter]

        return reduce(lambda a, b: a.append(b), wfs)


class PiecewiseConstant(WaveformPrimitive):
    def __init__(
        self,
        durations: List[ScalarType],
        values: List[ScalarType],
        parent: Optional[Builder] = None,
    ) -> None:
        assert len(durations) == len(
            values
        ), "durations and values must be the same length"
        super().__init__(parent)
        self._durations = list(map(ir.cast, durations))
        self._values = list(map(ir.cast, values))

    def __bloqade_ir__(self):
        iter = zip(self._values, self._durations)
        wfs = [ir.Constant(value=v, duration=t) for v, t in iter]
        return reduce(lambda a, b: a.append(b), wfs)


class Fn(WaveformPrimitive):
    def __init__(
        self,
        fn: Callable,
        duration: ScalarType,
        parent: Optional[Builder] = None,
    ) -> None:
        super().__init__(parent)
        self._fn = fn
        self._duration = ir.cast(duration)

    @beartype
    def sample(
        self, dt: ScalarType, interpolation: Union[ir.Interpolation, str, None] = None
    ):
        return Sample(dt, interpolation, self)

    def __bloqade_ir__(self):
        return ir.PythonFn.create(self._fn, self._duration)


# NOTE: no double-slice or double-record
class Slice(Waveform, Recordable):
    def __init__(
        self,
        start: Optional[ScalarType] = None,
        stop: Optional[ScalarType] = None,
        parent: Optional[Builder] = None,
    ) -> None:
        super().__init__(parent)
        # NOTE: this should no raise for None
        self._start = ir.scalar.trycast(start)
        self._stop = ir.scalar.trycast(stop)


class Record(Waveform, Sliceable):  # record should not be sliceable
    def __init__(
        self,
        name: str,
        parent: Optional[Builder] = None,
    ) -> None:
        super().__init__(parent)
        self._name = ir.var(name)


class Sample(Sliceable, Recordable, WaveformRoute):
    def __init__(
        self,
        dt: ScalarType,
        interpolation: Union[ir.Interpolation, str, None],
        parent: Builder,
    ) -> None:
        super().__init__(parent)
        self._dt = ir.cast(dt)
        if interpolation is None:
            self._interpolation = None
        else:
            self._interpolation = ir.Interpolation(interpolation)



from bloqade import piecewise_linear, start, var, cast
from bloqade.atom_arrangement import Chain
from bloqade.ir import (
    rydberg,
    detuning,
    AnalogCircuit,
    Sequence,
    Pulse,
    Field,
    AssignedRunTimeVector,
    Uniform,
)
import bloqade.ir.control.waveform as waveform
import bloqade.ir.scalar as scalar
from bloqade.codegen.common.assign_variables import AssignAnalogCircuit
from bloqade.ir.analysis.assignment_scan import AssignmentScan
from decimal import Decimal
import pytest


def test_assignment():
    lattice = Chain(2, 4.5)
    circuit = (
        lattice.rydberg.detuning.scale("amp")
        .piecewise_linear([0.1, 0.5, 0.1], [1.0, 2.0, 3.0, 4.0])
        .parse_circuit()
    )

    amp = 2 * [Decimal("1.0")]
    circuit = AssignAnalogCircuit(dict(amp=amp)).visit(circuit)

    target_circuit = AnalogCircuit(
        lattice,
        Sequence(
            {
                rydberg: Pulse(
                    {
                        detuning: Field(
                            drives={
                                AssignedRunTimeVector("amp", amp): piecewise_linear(
                                    [0.1, 0.5, 0.1], [1.0, 2.0, 3.0, 4.0]
                                )
                            }
                        ),
                    }
                )
            }
        ),
    )

    assert circuit == target_circuit


def test_assignment_error():
    lattice = Chain(2, 4.5)
    circuit = (
        lattice.rydberg.detuning.scale("amp")
        .piecewise_linear([0.1, 0.5, 0.1], [1.0, 2.0, 3.0, 4.0])
        .parse_circuit()
    )

    amp = 2 * [Decimal("1.0")]
    circuit = AssignAnalogCircuit(dict(amp=amp)).visit(circuit)
    with pytest.raises(ValueError):
        circuit = AssignAnalogCircuit(dict(amp=amp)).visit(circuit)


def test_scan():
    t = var("t")
    circuit = (
        start.rydberg.detuning.uniform.constant("max", 1.0)
        .slice(0, t)
        .record("detuning")
        .linear("detuning", 0, 1.0 - t)
        .parse_sequence()
    )

    params = dict(max=10, t=0.1)

    completed_params = AssignmentScan(params).emit(circuit)
    completed_circuit = AssignAnalogCircuit(completed_params).visit(circuit)

    t_assigned = scalar.AssignedVariable("t", 0.1)
    max_assigned = scalar.AssignedVariable("max", 10)
    detuning_assigned = scalar.AssignedVariable("detuning", 10)
    dur_assigned = 1 - t_assigned

    interval = waveform.Interval(cast(0), t_assigned)

    target_circuit = Sequence(
        {
            rydberg: Pulse(
                {
                    detuning: Field(
                        drives={
                            Uniform: waveform.Append(
                                [
                                    waveform.Slice(
                                        waveform.Constant(max_assigned, cast(1.0)),
                                        interval,
                                    ),
                                    waveform.Linear(detuning_assigned, 0, dur_assigned),
                                ]
                            )
                        }
                    )
                }
            )
        }
    )

    print(repr(completed_circuit))
    print(repr(target_circuit))

    assert completed_circuit == target_circuit



from dataclasses import dataclass
from numpy.typing import NDArray
from typing import TYPE_CHECKING
import numpy as np
from enum import Enum

if TYPE_CHECKING:
    from .emulator import Register
    from .atom_type import AtomType

MAX_PRINT_SIZE = 30


class SpaceType(str, Enum):
    FullSpace = "full_space"
    SubSpace = "sub_space"


@dataclass(frozen=True)
class Space:
    space_type: SpaceType
    atom_type: "AtomType"
    geometry: "Register"
    configurations: NDArray

    @classmethod
    def create(cls, register: "Register"):
        sites = register.sites
        n_atom = len(sites)
        atom_type = register.atom_type
        blockade_radius = register.blockade_radius
        Ns = atom_type.n_level**n_atom

        check_atoms = []

        for index_1, site_1 in enumerate(sites[1:], 1):
            site_1 = np.asarray(site_1)
            atoms = []
            for index_2, site_2 in enumerate(sites[:index_1]):
                site_2 = np.asarray(site_2)
                if np.linalg.norm(site_1 - site_2) <= blockade_radius:
                    atoms.append(index_2)

            check_atoms.append(atoms)

        min_int_type = np.min_scalar_type(Ns - 1)
        config_type = np.result_type(min_int_type, np.uint32)

        if all(len(sub_list) == 0 for sub_list in check_atoms):
            # default to 32 bit if smaller than 32 bit
            configurations = np.arange(Ns, dtype=config_type)
            return Space(SpaceType.FullSpace, atom_type, sites, configurations)

        states = np.arange(atom_type.n_level, dtype=config_type)
        configurations = states

        for index_1, indices in enumerate(check_atoms, 1):
            # assume no configurations are blocked
            mask = np.ones_like(configurations, dtype=bool)

            # loop over neighbors within blockade radius
            # find all non-blockaded configurations, e.g.
            # places where neighbors are in the ground state.
            # if the list if empty all configruations are
            # valid and should be included.
            for index_2 in indices:
                is_ground = np.logical_not(
                    atom_type.is_rydberg_at(configurations, index_2)
                )
                np.logical_and(is_ground, mask, out=mask)

            non_blockaded = configurations[mask]
            np.logical_not(mask, out=mask)
            blockaded = configurations[mask]

            # add new configurations
            # add all configurations because none of the configs are blockading
            new_non_blockaded = np.kron(non_blockaded, np.ones_like(states)) + np.kron(
                np.ones_like(non_blockaded), states * atom_type.n_level**index_1
            )
            # add all but the rydberg state because some at least one is blockading
            new_blockaded = np.kron(blockaded, np.ones_like(states[:-1])) + np.kron(
                np.ones_like(blockaded), states[:-1] * atom_type.n_level**index_1
            )

            configurations = np.hstack((new_blockaded, new_non_blockaded))

        configurations.sort()

        return cls(SpaceType.SubSpace, atom_type, sites, configurations)

    @property
    def index_type(self) -> np.dtype:
        if self.size < np.iinfo(np.int32).max:
            return np.int32
        else:
            return np.int64

    @property
    def size(self) -> int:
        return self.configurations.size

    @property
    def n_atoms(self) -> int:
        return len(self.geometry)

    @property
    def state_type(self) -> np.dtype:
        return self.configurations.dtype

    def is_rydberg_at(self, index: int) -> NDArray:
        return self.atom_type.is_rydberg_at(self.configurations, index)

    def is_state_at(self, index: int, state: int):
        return self.atom_type.is_state_at(self.configurations, index, state)

    def swap_state_at(self, index: int, state_1: int, state_2: int) -> NDArray:
        row_indices, col_config = self.atom_type.swap_state_at(
            self.configurations, index, state_1, state_2
        )

        if not isinstance(row_indices, slice):
            row_indices = np.argwhere(row_indices).ravel()

        if self.space_type is SpaceType.FullSpace:
            return (row_indices, col_config)
        else:
            col_indices = np.searchsorted(self.configurations, col_config)
            mask = col_indices < self.size
            mask[mask] = col_config[mask] == self.configurations[col_indices[mask]]

            if not np.all(mask):
                if isinstance(row_indices, slice):
                    row_indices = np.arange(self.size)

                return row_indices[mask], col_indices[mask]
            else:
                return row_indices, col_indices

    def transition_state_at(self, index: int, fro: int, to: int) -> NDArray:
        row_indices, col_config = self.atom_type.transition_state_at(
            self.configurations, index, fro, to
        )

        if not isinstance(row_indices, slice):
            row_indices = np.argwhere(row_indices).ravel()

        if self.space_type is SpaceType.FullSpace:
            return (row_indices, col_config)
        else:
            col_indices = np.searchsorted(self.configurations, col_config)

            mask = col_indices < self.size
            col_indices = col_indices[mask]
            row_indices = row_indices[mask]

            mask = col_config[mask] == self.configurations[col_indices]
            col_indices = col_indices[mask]
            row_indices = row_indices[mask]

            return (row_indices, col_indices)

    def fock_state_to_index(self, fock_state: str) -> int:
        state_int = self.atom_type.string_to_integer(fock_state)
        if self.space_type is SpaceType.FullSpace:
            return state_int
        else:
            index = np.searchsorted(self.configurations, state_int)
            if index >= self.size or state_int != self.configurations[index]:
                raise ValueError(
                    "state: {fock_state} not in rydberg blockade subspace."
                )

            return index

    def index_to_fock_state(self, index: int) -> str:
        if index < 0 or index >= self.size:
            raise ValueError(f"index: {index} out of bounds.")

        if self.space_type is SpaceType.FullSpace:
            return self.atom_type.integer_to_string(index, self.n_atoms)
        else:
            return self.atom_type.integer_to_string(
                self.configurations[index], self.n_atoms
            )

    def zero_state(self, dtype=np.float64) -> NDArray:
        state = np.zeros(self.size, dtype=dtype)
        state[0] = 1.0
        return state

    def sample_state_vector(
        self, state_vector: NDArray, n_samples: int, project_hyperfine: bool = True
    ) -> NDArray:
        from .atom_type import ThreeLevelAtomType

        p = np.abs(state_vector) ** 2
        sampled_configs = np.random.choice(self.configurations, size=n_samples, p=p)

        sample_fock_states = np.empty((n_samples, self.n_atoms), dtype=np.uint8)

        for i in range(self.n_atoms):
            sample_fock_states[:, i] = sampled_configs % self.atom_type.n_level
            sampled_configs //= self.atom_type.n_level

        if project_hyperfine and isinstance(self.atom_type, ThreeLevelAtomType):
            sample_fock_states[sample_fock_states == 1] = 0
            sample_fock_states[sample_fock_states == 2] = 1

        return sample_fock_states

    def __str__(self):
        # TODO: update this to use unicode
        output = ""

        n_digits = len(str(self.size - 1))
        fmt = "{{index: >{}d}}. {{fock_state:s}}\n".format(n_digits)
        if self.size < MAX_PRINT_SIZE:
            for index, state_int in enumerate(self.configurations):
                fock_state = self.atom_type.integer_to_string(state_int, self.n_atoms)
                output = output + fmt.format(index=index, fock_state=fock_state)

        else:
            lower_index = MAX_PRINT_SIZE // 2 + (MAX_PRINT_SIZE % 2)
            upper_index = self.size - MAX_PRINT_SIZE // 2

            for index, state_int in enumerate(self.configurations[:lower_index]):
                fock_state = self.atom_type.integer_to_string(state_int, self.n_atoms)
                output = output + fmt.format(index=index, fock_state=fock_state)

            output += (n_digits * "  ") + "...\n"

            for index, state_int in enumerate(
                self.configurations[upper_index:], start=self.size - MAX_PRINT_SIZE // 2
            ):
                fock_state = self.atom_type.integer_to_string(state_int, self.n_atoms)
                output = output + fmt.format(index=index, fock_state=fock_state)

        return output



import os
from typing import Optional


def sso_login(profile: Optional[str] = None):
    """
    checks if aws has log-in, and performs attempt of sso login if not.
    by default profile should be set as env variable `AWS_PROFILE`,
    but can be specified as a parameter.
    """
    if profile is not None:
        os.environ["AWS_PROFILE"] = profile
    code = os.system("aws sts get-caller-identity ")
    if code != 0:
        print("login rsp code is " + str(code) + ". attempting to sso login:")
        os.system("aws sso login")
